//----------------------------------------------------------------------------
// Anti-Grain Geometry - Version 2.3
// Copyright (C) 2002-2005 Maxim Shemanarev (http://www.antigrain.com)
//
// Permission to copy, use, modify, sell and distribute this software 
// is granted provided this copyright notice appears in all copies. 
// This software is provided "as is" without express or implied
// warranty, and with no claim as to its suitability for any purpose.
//
//----------------------------------------------------------------------------
// Contact: mcseem@antigrain.com
//          mcseemagg@yahoo.com
//          http://www.antigrain.com
//----------------------------------------------------------------------------
//
// Object that deliveres information about length, position and direction
// along a path (used by svg::text_path and line markers)
//
//----------------------------------------------------------------------------

#ifndef SVG_CONV_LOCATOR_INCLUDED
#include "agg_conv_curve.h"
#include "agg_conv_contour.h"
#include "agg_conv_stroke.h"
#include "agg_conv_transform.h"

#include "svg_basics.h"
#include "svg_vertex_source.h"
#include "svg_segment.h"
#define SVG_CONV_LOCATOR_INCLUDED

namespace svg
{
	//----------------------------------------------------------------------
	// This object treats a curve as sequence of short, straight, linear 
	// segments. It maintains on it a "current position", defined by the 
	// distance from the begin of the curve to the current position
	//----------------------------------------------------------------------
	class conv_locator
	{
	public:
		//Constructor
		//Arguments:
		//	vs:			source of the
		//	path_id:	id of the path from the vs
		//	trans:		global transformation, that has to be applied to vs
		conv_locator(vsource &vs, handle path_id, const transform &trans): 
		m_spos(0.0),
		m_path_id(path_id),
		m_trans(trans),
		m_src(vs),
		m_curve(m_src),
		m_vs(m_curve, m_trans) //m_c_curve(m_curve,m_trans),
		//m_vs(m_c_curve) 
		{
			m_size = -1.0; //in this way we mark that the calculation 
						   //is needed 
			m_continue = is_terminated;
			reset();
		}

		//Returns actual length form the begin to the actual position 
		//on the curve
		double position () {return m_spos + m_l0;}

		//Returns coordinates of the actual position on the curve
		//Arguments:
		//	x:	address of the x coordinate
		//	y:	address of the y coordinate
		void position(double *x, double *y){*x = m_a.x(m_l0); *y = m_a.y(m_l0);}

		//Advances actual position by given distance
		//Arguments:
		//	advance:	length that should be advanced
		//Returns false, when the end of the curve has been reached
		//(in this case most probably the real distance is shorter)
		bool move(double advance);

		//Resets the actual position on the curve to its begin
		void reset()
		{
			if (m_continue != is_closed) //closed curves are already 
			{							 //rewinded and positioned -
				m_vs.rewind(m_path_id); //(- see next_segment(bool))
				m_a = segment(0.0,0.0); //degenerated segment (length = 0)
										//as the start one...
			}
			m_spos = 0.0;
			m_continue = has_next;
			next_segment(); //initialize: m_l0, m_a
		}

		//Sets the current position on given distance from the begining
		//of the curve.
		//Arguments:
		//	distance:	requested position on the curve
		//Returns true, when the position has been successfully reached
		bool go_to(double distance) 
		{
			reset(); 
			return move(distance);
		}

		//Returns true, when the end of this curve has been reached
		bool terminated() 
		{	
			return (m_continue == is_terminated) 
					|| 
				   ((m_continue == is_closed) && (m_l0 >= m_a.length()));
		}


		//Returns the total length of the curve
		double length()
		{
			if(m_size < 0)
			{
				double prev = position();
				while(next_segment()); //scroll to the end;
				go_to(prev); //return to the previous position
			}
			return m_size;
		}

	protected:
		//Advances the current position to the begining of next 
		//segment of the curve.
		//Arguments:
		//	broken: true, when there was a gap before the selected
		//	segment.
		//Returns false, when the last segment has been reached.
		bool next_segment(bool &broken);

		//Simpler version: no information about broken
		//Returns false, when the last segment has been reached.
		bool next_segment() 
		{
			bool broken;
			return next_segment(broken);
		}

	protected:
		double		m_spos; //distance from the begining to 
							//the actual segment;
		double		m_l0;	//distance within actual segment to 
							//the actual position;
		segment		m_a;	//actual segment
		double		m_size;	//total length of the m_vs path
		handle		m_path_id; //id of the path
		enum stop_flag
		{
			has_next		= -1,   //curve has next segment:
			is_terminated	= 0,     //the end of the curve has been reached
			is_closed		= 1,	//for closed curves: we are on the closing
									//segment!
		}
		m_continue;	//flag, that detects the end of the curve

	//Curve "pipeline":
    protected:
		typedef agg::conv_curve<vertex_source>		curved;
		typedef agg::conv_transform<curved>			c_curved;
		//typedef agg::conv_contour<c_curved>			contour; 

		vertex_source	m_src;
		transform		m_trans; //actual transformation 
		
		curved			m_curve;
		c_curved		m_vs; //m_c_curve;
	};
}

#endif SVG_CONV_LOCATOR_INCLUDED
