//----------------------------------------------------------------------------
// SVG Viewer - Version 1.0
// A library of SVG renderer
// Copyright (C) 2008 Witold Jaworski
// Contact:	witold.jaworski@airplanes3d.net
//			http://airplanes3d.net
// 
// Contains elements from:
//			AGG - a graphic library (Copyright Maxim Shemanarev, mcseem@antigrain.com)
//          http://antigrain.com
// Uses functions from:
//          Expat - a XML parser library (Copyright James Clark)
// 
// SVG Viewer is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
// 
// SVG Viewer is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with SVG Viewer; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, 
// MA 02110-1301, USA.
//----------------------------------------------------------------------------

#ifndef SVG_FILTER_INCLUDED
#include "svg_shape.h"
#include "svg_req_render.h"
#define SVG_FILTER_INCLUDED

namespace svg
{

//////////////////////////////////////////////////////////////////////
// filter: container of the filters
//				
//////////////////////////////////////////////////////////////////////
class filter : public element 
{
	public:
		//////////////////////////////////////////////////////////////////////
		// constructor and overriden virtual methods
		//////////////////////////////////////////////////////////////////////

		filter(element *parent, unsigned level): element(parent, level),
		m_href(invalid_handle),
		m_x(0), m_y(0),
		m_width(0), m_height(0),
		m_relative_units (-1), //not initialized
		m_relative_content_units(-1), //not initialized, yet
		m_initialized(false)
		{}

		virtual void on_attribute(parser &prs, const char * name, const char *value)
		{
			if (strcmp(name, "filterUnits") == 0) 
			{	
				if (strcmp(value, "objectBoundingBox")==0) m_relative_units = 1;
				if (strcmp(value, "userSpaceOnUse")==0) m_relative_units = 0;
			}
			else
			if (strcmp(name, "primitiveUnits") == 0) 
			{	
				if (strcmp(value, "objectBoundingBox")==0) m_relative_content_units = 1;
				if (strcmp(value, "userSpaceOnUse")==0) m_relative_content_units = 0;
			}
			else
			if (strcmp(name, "xlink:href") == 0) 
			{	
				m_href  = prs.get_id(value); //id of another
			}
			else 
			if (strcmp(name, "x") == 0) 
			{	
				m_x = parse_coordinate(prs,value);
			}
			else
			if (strcmp(name, "y") == 0) 
			{	
				m_y = parse_coordinate(prs,value);
			}
			else
			if (strcmp(name, "width") == 0) 
			{	
				m_width = parse_coordinate(prs,value);
			}
			else
			if (strcmp(name, "height") == 0) 
			{	
				m_height = parse_coordinate(prs,value);
			}
			else
			{//if the style is not recognized:
				element::on_attribute(prs,name,value);
			}
		}

		virtual const char * tag() {return "filter";}; 

		//request handler: the override exists - see svg_filter.cpp
		virtual handle on_request(request &req); 
		
		virtual bool is_of(const char *cls) 
		{
			if(strcmp("filter",cls) == 0)
			{
				return true;
			}
			else //it is not your name? Make sure, asking your parent :)
				return element::is_of(cls);
		}

	public:
		//Renders the given elmement onto the result buffer, pointed by req
		//Arguments:
		//	elm:	element, which picture will be filtered and applied to the 
		//			result;
		//	req:	actual request to render 
		//	opacity:optional - level of transparency, used to apply filter
		//			effect to the result buffer (the buffer pointed in the
		//			the <req> request)
		void perform(shape &elm, req_render &req, 
											agg::int8u opacity = agg::cover_full);
		
		// Determines, how bigger will be the intermediate buffers comparing to 
		// the target area (some extent is required for example for the gaussian 
		// blurs - otherwise they would be distored at the edges). Default: 2.0 
		// (50% from each side). It may be override in the child classes. Used in 
		// the filter::preform() method.
		virtual double extent_ratio() {return 2.0;}
		//////////////////////////////////////////////////////////////////////
		// Simple methods for copying the content of the referenced filter
		// (see apply() method)
		//////////////////////////////////////////////////////////////////////
		int relative_units() {return m_relative_units;}

		int relative_content_units() {return m_relative_content_units;}

		//raw region of the filter (to copy setting from
		rectangle region() {return rectangle(m_x, m_y, m_x+m_width, m_y+m_height);}

		bool initialized() {return m_initialized;}

		//////////////////////////////////////////////////////////////////////
		// Template used to copy data between intermediate picture buffers
		//////////////////////////////////////////////////////////////////////
		//Copies whole buffer src into specified place on dst
		//Arguments:
		//	dst:	destination byffer
		//	src:	source buffer
		//	x:		distance from left edge of dst to first pixel, placed from
		//			src;
		//	y:		distance from top edge of dst to first pixel, placed from
		//			src;
		//	clip:	optional - address to an additional clip box (it can be 
		//			sometimes smaller than ths size of dst buffer)
		//  opacity:optional - overall opacity, that shoul be applied to the
		//			source (incoming pixels can be transparent)
        template<class DstPixFmt, class SrcPixFmt> 
		static void copy(DstPixFmt &dst, SrcPixFmt &src, int x, int y, 
							agg::rect_i *bounds = NULL, agg::int8u opacity = 255)
		{
			agg::rect_i clp, sbox, dbox;

			//clip input areas:
			sbox = agg::rect_i(x, y, x+src.width()-1,y+src.height()-1);
			dbox = agg::rect_i(0, 0, dst.width()-1,dst.height()-1);

			if (bounds) //let's do intersection (to ensure, that 
			{			//bounds do not execeed the dst region):
				dbox = agg::intersect_rectangles(*bounds,dbox);
			}

			clp = agg::intersect_rectangles(sbox, dbox);
			//calculate valid ranges:
			if (clp.is_valid())
			{
				unsigned dx = clp.x1; 
				unsigned dy = clp.y1;
				unsigned sx = clp.x1 - x;
				unsigned sy = clp.y1 - y;
				unsigned width = clp.x2 - clp.x1 + 1;
				unsigned height = clp.y2 - clp.y1 + 1;
				//perform:
				unsigned row;
				for(row = 0; row < height; row++)
				{
					dst.blend_from(src,dx,dy+row,sx,sy+row,width,opacity);
				}
			}
		}//copy
				

	protected:
		//Helper function: changes percent coordinates into simple fractions
		//(it is OK for this specific case - gradients)
		//Arguments:
		//	prs:	the parser
		//	value:	text, that will be interpreted as coordinate (may be 
		//			relative, in %)
		static double parse_coordinate(parser &prs, const char *value)
		{
			bool percent = false;
			return prs.to_length(value, &percent);
		}

		//////////////////////////////////////////////////////////////////////
		// New virtual methods
		//////////////////////////////////////////////////////////////////////
		//Copies settings from another gradient. Called within initialize() method 
		//Arguments:
		//	req:	a request object (request::initialize)
		//	src:	another filter - source of the information (of the same class!)
		//This method exists, because in SVG one filter element can 
		//reference another, and the sum of the attributes from the BOTH gives 
		//the all information required to create a filter
		virtual void use(request &req, filter *src)
		{
			if(m_relative_units < 0 && src->relative_units() >= 0) //initialized? 
			{
				m_relative_units = src->relative_units();
			}

			if(m_relative_content_units < 0 && src->relative_content_units() >= 0) //initialized? 
			{
				m_relative_content_units = src->relative_content_units();
			}

			if (has_no_area(region()) && !has_no_area(src->region()))
			{
				rectangle sr = src->region();
				m_x = sr.x1;
				m_y = sr.y1;
				m_width = sr.x2 - sr.x1;
				m_height = sr.y2 - sr.y1;
			}

		}

		//Initializes the filter, when the object receives a 
		//reqest::initialize
		//Arguments:
		//	req:	req::initialize - passed to give the access to the element and
		//			attribute storage.	
		virtual void initialize(request &req)
		{
			if (initialized()) return; //nothing to do
			filter *reference = (filter *) req.get_element(m_href);
			if (reference) 
			{
				//give it a chance to initialze itself, before it 
				//initializes the others (it sometimes may happen, that 
				//the object referenced is after the reference user!):
				if (!reference->initialized()) reference->on_request(req);
				//OK, initialize it!
				use(req, reference);
			}
			
			if (m_relative_units < 0) 
			{					      //default:	
				m_relative_units = 1; //filterUnits = "objectBoundingBox"
			}

			if (m_relative_content_units < 0) 
			{								  //default:	
				m_relative_content_units = 0; //primitiveUnits = "userSpaceOnUse"
			}
			
			if(has_no_area(region()))//if no coordinates:
			{	// assume the default ones (according SVG spec):
				m_x = -0.1;
				m_y = -0.1;
				m_width = 1.2;
				m_height = 1.2;

				m_relative_units = 1; //filterUnits = "objectBoundingBox"
			}

			//Give the chance to the children elements to initalize
			element::forward(req);

			m_initialized  = true;
		}

		//////////////////////////////////////////////////////////////////////
		// Other helper methods
		//////////////////////////////////////////////////////////////////////
		
	protected:
		//////////////////////////////////////////////////////////////////////
		// Instance additional fields
		//////////////////////////////////////////////////////////////////////
		handle		m_href;	   //optional: reference to another filter definition
							   //which attributes can override these ones.
							   //Can be = invalid_handle

		//flags
		//filterUnits:
		int	m_relative_units;//tristate: -1, when not initialized, 
							 //0: when userSpaceOnUse is declared
							 //1: when the objectBoundingBox is declared

		//primitiveUnits:
		int	m_relative_content_units; //tristate: -1, when not initialized, 
									 //0: when userSpaceOnUse is declared
									 //1: when the objectBoundingBox is declared

		bool	m_initialized;	//a flag, that allows to check if the initialized()
								//on this object was called at least once.

		//"Raw" information, as it was read from the XML source:
		// The coordinates are in the nominal object units, like the shape
		// bounding box, and need to be transformed to the relative units 
		// (if filterUnits = userSpaceOnUse). Then they are applied to
		// TRANSFROMED shape (expressed in the coordinates of the result picture.
		double m_x, m_y;			//upper - left corner of the filter region
		double m_width, m_height;	//width and height of the filter region

}; //filter

}; //namespace svg
#endif SVG_FILTER_INCLUDED
