//----------------------------------------------------------------------------
// SVG Viewer - Version 1.0
// A library of SVG renderer
// Copyright (C) 2008 Witold Jaworski
// Contact:	witold.jaworski@airplanes3d.net
//			http://airplanes3d.net
// 
// Contains elements from:
//			AGG - a graphic library (Copyright Maxim Shemanarev, mcseem@antigrain.com)
//          http://antigrain.com
// Uses functions from:
//          Expat - a XML parser library (Copyright James Clark)
// 
// SVG Viewer is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
// 
// SVG Viewer is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with SVG Viewer; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, 
// MA 02110-1301, USA.
//----------------------------------------------------------------------------
#include <stdlib.h>
#include "svg_exception.h"
#include "svg_font.h"

namespace svg
{
	static char *def_size = "72px"; //default font size (let it be large...)

	//------------------------------------------------------------------------
	//Helper function, used to recognize the elements of text font spec:
	//font-style, font-variant, font-weight.
	//Arguments
	//	token:	element of the font specification (the font style element)
	//			it is the subject of this recoginition
	//	italic: set to true, when the token is recognized as the font-style
	//			value
	//	width:	width of the font: a value from 100 (lighter) to 900 (bolder) 
	//
	//	Function returns true, when it has recognized the token
	static bool recognize(const char *token, bool *italic, 
												font::weight_value *weight)
	{
		//"normal" can apply to all three recognized elements
		//just leave it alone:
		if (strcmp(token, "normal") == 0) return true;
		
		//font-style?:
		if (
				strcmp(token, "italic") == 0
				||
				strcmp(token, "oblique") == 0
			)
		{	*italic = true;		return true;	}

		//font-variant?
		if (strcmp(token, "small-caps") == 0) return true;

		//font-weight?:
		if (strcmp(token,"100") == 0)
			{	*weight = font::w100;		return true;	}
		if (strcmp(token,"200") == 0)
			{	*weight = font::w200;		return true;	}
		if (strcmp(token,"300") == 0)
			{	*weight = font::w300;		return true;	}
		if ((strcmp(token,"400") == 0))
			{	*weight = font::w400;		return true;	}
		if (strcmp(token,"500") == 0)
			{	*weight = font::w500;		return true;	}
		if (strcmp(token,"600") == 0)
			{	*weight = font::w600;		return true;	}
		if (
			(strcmp(token,"700") == 0)
			||
			(strcmp(token,"bold") == 0)
		   )
			{	*weight = font::w700;		return true;	}
		if (strcmp(token,"800") == 0)
			{	*weight = font::w800;		return true;	}
		if (strcmp(token,"900") == 0)
			{	*weight = font::w900;		return true;	}

		return false; //token unrecognized
	}//recognize
	//------------------------------------------------------------------------
	//Initializes the font, using the font declaration string
	//Arguments:
	//	spec:	font specificaton
	void font::initialize(const char *spec)
	{
		bool				italic	= false;
		font::weight_value	weight	= font::normal; 
		char				value[SVG_MAX_NAME]; //a buffer for a single 
												 //value from the spec
		char *p = (char *) spec;
		strcpy(value,"normal"); //"normal" is neutral in the recognition

		while (*p && recognize(value,&italic,&weight))
		{
			p = parser::get_first_token(p,' ',value, sizeof(value));
		}

		//the not recognized value in the specification is the font size;
		//it is already loaded into value buffer, so the p points to the
		//next, the last element - font family (which can contain some spaces!)
		initialize(p,italic,weight); //call the virtual function, from the 
									 //child class
	}
	//------------------------------------------------------------------------
	//Helper function, used in the two font::specification() methods, below
	//"Glues" a string from given text components array
	//Arguments:
	//	buf:		a buffer that will be filled
	//	bufsize:	buffer capacity;
	//	src:		an array of pointers to text components, that have to be glued.
	//				when a component is NULL or "", it is skipped
	//	count:		number of elements in the src array;
	//	separator:  separator, which should be placed in the buffer between 
	//				the glued components
	//	Returns NULL, when the components have not fitted successfully into buf.
	//	Otherwise - the address of the bif.
	static char * glue(char *buf, size_t bufsize, 
								char**src, unsigned count, char separator = ' ')
	{
		unsigned i;
		char *p = buf; 
		bufsize--;	
		p[bufsize] = 0; //add the railing null - just to be sure...
		
		for(i = 0; i < count && bufsize > 0; i++)
		{
			if ((i> 0) && (p > buf) && (p[-1] != separator)) //add separator:
			{
				*p = separator; p++;
				bufsize --;
			}
			
			if (src[i] && *(src[i])) //it is non-null and non-empty?
			{
				strncpy(p, src[i], bufsize);
				bufsize -= strlen(src[i]);
				p		+= strlen(src[i]);
			}
		}//for

		if(bufsize > 0) 
		{
			if ((p > buf) && (p[-1] == separator)) p--; //"eat" the trailing separator
			*p = 0; //add trailing null
			return buf;
		}
		else
		{
			return NULL;
		}

	}//glue
	//------------------------------------------------------------------------
	//Helper function: creates the font specification from separate
	//attribute values.
	//Arguments:
	//	buf:	a buffer for the specification text
	//	bufsize:buffer capacity
	//	family:	font family (for example: "Arial", "Times New Roman")
	//	style:	"italic" or "normal"
	//	weight: "100","200","300",.."900" or "normal", or "bold".
	//	parent:	an SVG graph element used to obtain information
	//			about any missing attribute, from the parent element's style
	char *font::specification(char *buf, size_t bufsize, char *family, 
									 char *style, char *weight, element *parent)
	{
		if(!*family || !*style || !*weight) //Part of the spec is missing:
		{									//we have ask the parent about actual font:
			req_get_style	req;
			parent->to_root(req); //get information about effective parent style

			//default font - defined in one of the the parent element, on the way to the root
			font	*def_font = (font *) req.get_element(req.style.of(attribute::font).val.id, "ext:font");

			if(!*family) //missing family?
			{
				strcpy(family, def_font? def_font->family() : "Serif"); 
			}
			if(!*style) //missing style?
			{
				strcpy(style, def_font && def_font->italic()? "italic" : "normal"); 
			}
			if(!*weight) //missing weight?
			{
				if(def_font)
				{
					strcpy(weight, "000"); //we will update this in the code below:
					switch(def_font->weight())
					{
					case font::w100:
						weight[0] = '1'; break;
					case font::w200:
						weight[0] = '2'; break;
					case font::w300:
						weight[0] = '3'; break;
					case font::normal:
						strcpy(weight,"normal");
						break;
					case font::w500:
						weight[0] = '5'; break;
					case font::w600:
						weight[0] = '6'; break;
					case font::bold:
						strcpy(weight,"bold");
						break;
					case font::w800:
						weight[0] = '8'; break;
					case font::w900:
						weight[0] = '9'; break;
					default:
						strcpy(weight,"normal");
					}//end switch
				}
				else
				{
					strcpy(weight,"normal");
				}//end if(def_font)
			}//end if(!*weight)
		}//end if (!*family || !*style || ...)
		//at this point all three text components of the specification:
		//family, style, weight, have a non-empty value. Now it is time 
		//to fill the result buffer:
		char *src[] = {style,weight,def_size,family}; //string sequence to
													  //be placed in the buffer
		//Glue it with the spaces!
		return glue(buf,bufsize,src,sizeof(src)/sizeof(src[0]),' ');
		
	}//specification
	//Called internally by the initialize(spec) function.
	// Arguments:
	//	buf:	a buffer for the specification text
	//	bufsize:buffer capacity
	//	family:	font family name, like "Arial" or "Times New Roman"
	//	italic: true, when the font has to in the italic version
	//	weight: weight of the font
	char * font::specification(char *buf, size_t bufsize, const char *family, 
											bool italic, font::weight_value weight)
	{
		char word[SVG_MAX_NAME]; 
		char *words[] = {NULL, NULL, def_size, (char *) family}; 

		if (italic) words[0] = "italic";
	
		if(weight != font::normal)
		{
			strcpy(word,"000"); //we will update the first character below:
			switch(weight)
			{
				case font::w100:
					word[0] = '1'; break;
				case font::w200:
					word[0] = '2'; break;
				case font::w300:
					word[0] = '3'; break;
				//		   w400: = normal
				case font::w500:
					word[0] = '5'; break;
				case font::w600:
					word[0] = '6'; break;
				case font::bold:
					strcpy(word,"bold");
					break;
				case font::w800:
					word[0] = '8'; break;
				case font::w900:
					word[0] = '9'; break;
					break;
			}//end switch
				
			words[1] = word;
		}

		//Glue it with the spaces!
		return glue(buf,bufsize,words,sizeof(words)/sizeof(words[0]),' ');
	}//specification


} // namespace svg