//----------------------------------------------------------------------------
// SVG Viewer - Version 1.0
// A library of SVG renderer
// Copyright (C) 2008 Witold Jaworski
// Contact:	witold.jaworski@airplanes3d.net
//			http://airplanes3d.net
// 
// Contains elements from:
//			AGG - a graphic library (Copyright Maxim Shemanarev, mcseem@antigrain.com)
//          http://antigrain.com
// Uses functions from:
//          Expat - a XML parser library (Copyright James Clark)
// 
// SVG Viewer is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
// 
// SVG Viewer is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with SVG Viewer; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, 
// MA 02110-1301, USA.
//----------------------------------------------------------------------------

#ifndef SVG_GRADIENT_INCLUDED
#include "agg_span_allocator.h"
#include "agg_span_interpolator_linear.h"
#include "agg_span_gradient.h"

#include "svg_element.h"
#include "svg_segment.h"
#include "svg_stop.h"
#include "svg_painter.h"

#define SVG_GRADIENT_INCLUDED

namespace svg
{
//----------------------------- Colors -------------------------------
//////////////////////////////////////////////////////////////////////
// color_segment: models a linear color transition form one to another
//				  through a "distance" of parameter values
//////////////////////////////////////////////////////////////////////
struct color_segment
{
	public: 
		//creates a new color segment, from two <stop> elements
		//Arguments:
		//	begin:	an element of the segment definition;
		//	end:	next element of the segment definition
		color_segment(stop &begin, stop &end)
		{
			m_offset = begin.offset();
			m_span = end.offset() - begin.offset();
			m_base = begin.color();
			m_end = end.color();
		}

		//Returns the start offset of the segment begin (0.0 ... 1.0)
		double	begin() {return m_offset;}

		//Returns the end offset of the segment (0.0 ... 1.0)
		double	end() {return m_offset + m_span;}

		//Return color for a given offset
		//Arguments:
		//	at:		position within segment (0.0 = begin, 1.0 = end)
		//Returns: a color at given parameter position
		rgba8	color(double at)
		{
			if (m_span < SVG_EPSILON) return m_end; //degenerated segement!
			double k = (at - m_offset)/m_span;
			rgba8 result = m_base.gradient(m_end, k);
			//if (opacity < 1.0) result.opacity(opacity * result.opacity());
			return result;
		}

	protected:
		double	m_span;		//span of the parameter values (0.0 .. 1.0)
		double	m_offset;	//offset of the begin color, in the parameter values
							//(0.0 .. 1.0)
		rgba8	m_base;		//the color of the begining
		rgba8	m_end;		//the color of the end
		
};//color_segment

//////////////////////////////////////////////////////////////////////
// color_generator: a base class that deliveres the gradient color values
//			 
//////////////////////////////////////////////////////////////////////
class color_array
{
	protected:

		enum cnst 
		{
			max_colors = 512 //maximum colors in the whole gradient span
		};

	public: 
		color_array(): m_set(false), m_opacity (1.0) 
		{ //clear the colors and their alphas - they are initially transparent
			memset(m_colors,0,sizeof(m_colors));
			memset(m_alpha,0,sizeof(m_alpha));
		}

		bool is_valid() {return m_set;}

		//Prepares the generator for the use:
		//Arguments:
		//	opacity:	(0.0 ... 1.0) - the overall opacity of the object
		void set(double opacity)
		{
			if (m_opacity != opacity)
			{
				unsigned i;
				for (i = 0; i < size(); i++)
				{
					m_colors[i].a = m_alpha[i]; //restore original opacity
					//multiple by the global opacity
					m_colors[i].opacity(m_colors[i].opacity()*opacity);
				}

				m_opacity = opacity;
			}
		}//end set

		//Returns the number of colors in this logical color array
		static unsigned size()	{return max_colors;}
		

	//Initializes a color array (inside a request::initialize handler):
	//	Arguments:
	//	req:		request - usually render::initialize;
	//	first_id:	id of the first <stop> element;
	//	count:		number of the elements;
	//				some of them can be of a different kind, than <stop>
	void initialize(request &req, handle first_id, int count)
	{
		m_opacity = 1.0; //opacity: full
		if (!svg::is_valid(first_id)) return; //nothing todo

		double total_length = (double) (size());
		double delta = 1.0 / total_length;
		
		//create the content of this color array
		stop *last, *first = find_stop(req, first_id, count);
		while(first && count > 0)
		{
			last = find_stop(req,first->id()+1,count); 
			if(last)
			{
				color_segment seg(*first, *last);
				unsigned i = 0;
				double k = seg.begin() + 0.5 * delta;
				while (k <= seg.end())
				{
					i = agg::ufloor(k*total_length);
					if (i < size()) //just to not overwrite something
					{				//by an incident!
						m_colors[i] = seg.color(k);
						m_alpha[i]	= m_colors[i].a;
					}
					k += delta;
				}//end while
			}//end if
			first = last;
		}//end for

		m_set = true;
	}

	//Returns a color at the given position (offset) v;
	//Arguments:
	//	v:	 offset to the value
	const rgba8 & operator [](unsigned v) const
	{
		return m_colors[v];
	}
	protected:
		//because in the gradient can appear an additional elements, 
		//that are absent in the source XML (in effect of early referencing)
		//We have to create a special funtion, which will return a next stop
		//element
		//Arguments:
		//	req:		request - usually render::initialize;
		//	id:			id of the first element to examine;
		//	count:		total number of remaining elements (updated in this call)
		//				some of them can be of a different kind, than <stop>
		//Function returns NULL, when no next element exists
		stop *find_stop(request &req, handle id, int &count)
		{
			stop * elm = NULL;
			while (elm == NULL && count > 0)
			{
				elm = (stop *) req.get_element(id, "stop");
				id++;
				count--;
			}
			return elm;
		}
	protected:
		double		m_opacity; //last used value of the "global opacity"
		rgba8		m_colors[max_colors];
		agg::int8u	m_alpha[max_colors]; //helper buffer for the opacities
		bool		m_set; //true, when it was initialized
};//color_array

//----------------------------- Offset -------------------------------
//////////////////////////////////////////////////////////////////////
// offset_generator: a base class that deliveres the gradient offset values
//			 
//////////////////////////////////////////////////////////////////////
//Pure virtual class - the base for all generators
class offset_generator
{
public:
	virtual int calculate(int x, int y, int d) const = 0;
};
//////////////////////////////////////////////////////////////////////
// gradient_clamp_adaptor: a class that implements the "pad" (no repeat)
//			               gradient extrapolation. This mode is required 
//						   by the SVG spec, and not implemented in the AGG
//////////////////////////////////////////////////////////////////////
template <class GradientFunction>class gradient_clamp_adaptor
{
	public:
		GradientFunction *m_gradient; //instance of agg::gradient_x, 
									  //or agg::gradient_radial
	public:
		gradient_clamp_adaptor(GradientFunction &gradient):
											  m_gradient(&gradient) {}
		AGG_INLINE int calculate(int x, int y, int d) const
		{
			int ret = m_gradient->calculate(x,y,d);
			if (ret < 0) ret = 0;
			if (ret > d) ret = d;
			return ret;
		}
};
//////////////////////////////////////////////////////////////////////
// polymorphic_offset_generator: a class that wraps around a selected
//			               offset_generator instance  
//						   
//////////////////////////////////////////////////////////////////////
class polymorphic_offset_generator: public offset_generator
{
	public:
		offset_generator *m_generator; //specific instance of a children
									  //class, that inherited from
									  //offset_generator
	public:
		polymorphic_offset_generator(){ m_generator = NULL; }

		//source: an instance of a particular generator assigned to this wrapper
		void source(offset_generator *generator) {m_generator = generator;}
		offset_generator *source() {return m_generator;}

		//returns true, when generator is ready to use (has received a source)
		bool is_valid() {return m_generator != NULL;}

		//virtual function from the base class:
		virtual int calculate(int x, int y, int d) const
		{
			return m_generator->calculate(x,y,d);
		}
};//polymorphic_offset_generator

//////////////////////////////////////////////////////////////////////
// gradient: a base class - element of the SVG graph, 
//			 which define a gradient for fillin or stroking
//////////////////////////////////////////////////////////////////////
class gradient : public painter 
{

	public:
		//////////////////////////////////////////////////////////////////////
		// constructor and overriden virtual methods
		//////////////////////////////////////////////////////////////////////

		gradient(element *parent, unsigned level): painter(parent, level),
		m_interpolator(m_actual_trans),
		//the gradient span - color_array.size() - should match the scale of 
		//m_internal_trans - see method initialize(), below.
		m_gradient(m_interpolator, m_offset_generator,m_color_generator,
														0.0,color_array::size())
		{}

		virtual void on_attribute(parser &prs, const char * name, const char *value)
		{
			if (strcmp(name, "gradientUnits") == 0) 
			{	
				if (strcmp(value, "objectBoundingBox")==0) m_relative_units = 1;
				if (strcmp(value, "userSpaceOnUse")==0) m_relative_units = 0;
			}
			else
			if (strcmp(name, "gradientTransform") == 0) 
			{
				m_extra_trans = prs.parse_transform(value);
				m_extra_trans.invert();
			}
			else
			if (strcmp(name, "spreadMethod") == 0) 
			{	
				m_offset_generator.source(select_generator(value));
			}
			else 
			{//if the style is not recognized:
					painter::on_attribute(prs,name,value);
			}
		}
		
		virtual bool is_of(const char *cls) 
		{
			if(strcmp("gradient",cls) == 0)
			{
				return true;
			}
			else //it is not your name? Make sure, asking your parent :)
				return painter::is_of(cls);
		}

		//Method required by AGG to be used in the render_scanlines_aa() method:
		//to be overriden in the children classes:
		virtual void generate(rgba8* span, int x, int y, unsigned len)
		{
			m_gradient.generate(span, x, y, len);
		}

		//Call this method inside req_render::perform method,
		//just before the render_scanlines_aa
		//Arguments:
		//	gtrans:	"global" transformation of the shape (from its local 
		//			coordinates to the picture coordinates
		//	opacity:overall opacity of the gradient/pattern. (0.0 .. 1.0)
		//			used in the child classes
		virtual void set(transform &gtrans, double opacity)
		{
			//We have to set the default offset generator just before rendering
			// - if we would do it in the initialize(), the default setting
			//from the href-erenced gradient overrides the generator properly
			//set before, during xml - parsing stage.
			if(!m_offset_generator.is_valid())
			{ //pad: default way of fillig in!
				m_offset_generator.source(select_generator("pad"));
			}

			m_color_generator.set(opacity);
			painter::set(gtrans, opacity);
		}

	public:
		//////////////////////////////////////////////////////////////////////
		// Simple methods for copying the content of the referenced gradient
		// (see apply() method)
		//////////////////////////////////////////////////////////////////////
		//the id of the first stop
		handle stop_id() { return first_child();}
		
		segment span() {return m_span;}

		//the offset generator: it can be also copied
		polymorphic_offset_generator &offset_generator(){return m_offset_generator;} 


		//////////////////////////////////////////////////////////////////////
		// Other methods and fields
		//////////////////////////////////////////////////////////////////////

	protected:
		//This method is called within painter::initialize, if the m_href is valid.
		virtual void use(request &req, painter *src)
		{
			if (!src->is_of("gradient")) return; //something wrong - this is not a gradient

			gradient *href = (gradient *) src;
			
			if (!m_span.is_valid() && href->span().is_valid()) //copy span, if valid 
			{
				m_span = href->span();
			}

			if (!offset_generator().is_valid() && href->offset_generator().is_valid()) //copy generator, if valid
			{
				m_offset_generator.source(href->offset_generator().source());
			}
			
			if(elements() <= 1 && href->elements() > 1) //has stops, to create a color array?
			{
				m_color_generator.initialize(req, href->stop_id(), href->elements());
			}

			painter::use(req, src);
		}	
	

		//Initializes the gradient, when the object receives a reqest::initialize
		//Arguments:
		//	req:	req::initialize - passed to give the access to the element and
		//			attribute storage.	
		//
		// BEWARE: in the child classes initialize the m_span, before calling this one! 
		//
		virtual void initialize(request &req)
		{
			painter::initialize(req);
			if(!m_color_generator.is_valid()) 
			{
				m_color_generator.initialize(req, stop_id(), elements()); 
			}

			//form the internal transformation:
			m_internal_trans = transform();
			m_internal_trans.scale(1.0 / (double) color_array::size());

			if(m_span.is_valid())
			{
				m_internal_trans.scale(m_span.length());
				m_internal_trans.rotate(m_span.angle());
				m_internal_trans.translate(m_span.x0(), m_span.y0());
				m_internal_trans.invert();
			}
		}

		//Returns an offset generator object matching the given symbol
		//Arguments:
		//	symbol:	one of three values: "repeat", "reflect", "pad"
		//Each child class should return a correspodnig offset generator
		virtual svg::offset_generator *select_generator(const char *symbol)
		{
			return &m_offset_generator;
		}

		
	public:
		typedef agg::span_gradient<rgba8, 
									agg::span_interpolator_linear<>,
									polymorphic_offset_generator, 
									color_array>				span_generator;
	protected:
		//////////////////////////////////////////////////////////////////////
		// Instance additional fields
		//////////////////////////////////////////////////////////////////////
		segment		m_span;	   //for linear gradients: segment between x1,y1 and
							   //x2, y2. For radial gradients: horizontal segment
							   //of origin at cx,cy and end at cx+r, cy.
							   //useful for many internal calculations

		//Objects, that build the gradient "pipeline":

		//1. transforms the drawing x,y coorinates into "gradient coordinate
		//   system". This system has 0.0 point at the gradient origin, the axis
		//	 aligned to gradient direction, and the unit length equal the gradient
		//	 span.
		agg::span_interpolator_linear<> m_interpolator;

		//2. for gradient x,y coordinates, calculated in the previous step,
		//   returns returns an offset
		polymorphic_offset_generator m_offset_generator;

		//3. for an offset, calculated in the previous step, returns a color
		color_array m_color_generator; 

		//And finally: the gradient object itself:
		span_generator m_gradient;
}; //gradient

}; //namespace svg
#endif SVG_GRADIENT_INCLUDED
