//----------------------------------------------------------------------------
// SVG Viewer - Version 1.0
// A library of SVG renderer
// Copyright (C) 2008 Witold Jaworski
// Contact:	witold.jaworski@airplanes3d.net
//			http://airplanes3d.net
// 
// Contains elements from:
//			AGG - a graphic library (Copyright Maxim Shemanarev, mcseem@antigrain.com)
//          http://antigrain.com
// Uses functions from:
//          Expat - a XML parser library (Copyright James Clark)
// 
// SVG Viewer is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
// 
// SVG Viewer is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with SVG Viewer; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, 
// MA 02110-1301, USA.
//----------------------------------------------------------------------------

#ifndef SVG_PAINTER_INCLUDED
#include "agg_span_allocator.h"
#include "svg_element.h"
#include "svg_req_render.h"
#define SVG_PAINTER_INCLUDED

namespace svg
{
//----------------------------- Colors -------------------------------
//////////////////////////////////////////////////////////////////////
// painter: base class for both: gradient and pattern
//				
//////////////////////////////////////////////////////////////////////
class painter : public element 
{
	public:
		//////////////////////////////////////////////////////////////////////
		// constructor and overriden virtual methods
		//////////////////////////////////////////////////////////////////////

		painter(element *parent, unsigned level): element(parent, level),
		m_href(invalid_handle),
		m_extra_trans(0,0,0,0,0,0),
		m_relative_units (-1), //not initialized
		m_relative_content_units(-1), //not initialized, yet
		m_initialized(false)
		{}

		virtual void on_attribute(parser &prs, const char * name, const char *value)
		{
			if (strcmp(name, "xlink:href") == 0) 
			{	
				m_href  = prs.get_id(value); //id of another
			}
			else 
			{//if the style is not recognized:
					element::on_attribute(prs,name,value);
			}
		}


		//request handler:
		virtual handle on_request(request &req) 
		{
			switch(req.type())
			{
			case request::initialize:
					initialize(req);
					return skip(req);
			default: 
				return element::on_request(req);
			}//switch
		}
		
		virtual bool is_of(const char *cls) 
		{
			if(strcmp("painter",cls) == 0)
			{
				return true;
			}
			else //it is not your name? Make sure, asking your parent :)
				return element::is_of(cls);
		}

	public:

		//////////////////////////////////////////////////////////////////////
		// New virtual methods
		//////////////////////////////////////////////////////////////////////
		//Method required by AGG to be used in the render_scanlines_aa() method:
		//may be overriden in the children classes, actually not used
		virtual void prepare() {}

		//Method required by AGG to be used in the render_scanlines_aa() method:
		//to be overriden in the children classes:
		virtual void generate(rgba8* span, int x, int y, unsigned len) {}
		
		//Call this method inside req_render::perform method,
		//just before the render_scanlines_aa
		//Arguments:
		//	gtrans:	"global" transformation of the shape (from its local 
		//			coordinates to the picture coordinates
		//	opacity:overall opacity of the gradient/pattern. (0.0 .. 1.0)
		//			used in the child classes
		virtual void set(transform &gtrans, double opacity)
		{
			//form the final transformation from picture to internal coordinates:
			transform global = gtrans;
			global.invert();
			m_actual_trans = transform();
			m_actual_trans *= global;
			m_actual_trans *= m_extra_trans;
			m_actual_trans *= m_unit_trans;
			m_actual_trans *= m_internal_trans;
		}

		//Sets the unit transformation (when unit mode is objectBoundingBox)
		//Arguments:
		//	box:	the reference bounding box (use shape::bounding_box() here)
		//Should be called for gradients which have relative units, only.
		//Call it BEFORE req_render::perform method
		virtual void set(rectangle &box)
		{
			m_unit_trans = transform();
			//scale y is negative, because it points downwards:
			m_unit_trans.scale(box.x2 - box.x1, box.y1 - box.y2); 
			m_unit_trans.translate(box.x1, box.y1);
			m_unit_trans.invert();
		}

		//an alternative to set(box): called, when set(box) is NOT called
		virtual void reset()
		{
			m_unit_trans = transform();
		}
		//////////////////////////////////////////////////////////////////////
		// Simple methods for copying the content of the referenced gradient
		// (see apply() method)
		//////////////////////////////////////////////////////////////////////
		int relative_units() {return m_relative_units;}

		int relative_content_units() {return m_relative_content_units;}

		bool initialized() {return m_initialized;}
		
		//the additional transformation (usually - an identity)
		transform &extra_transform() {return m_extra_trans;}

		//////////////////////////////////////////////////////////////////////
		// Other methods and fields
		//////////////////////////////////////////////////////////////////////
		//flag used to signalize, that the set(box) method is required
		bool box_required() {return m_relative_units == 1 || m_relative_content_units == 1;}

		//Helper method to retrieve the painter object from a styleset
		//Arguments:
		//	style:	the queried styleset
		//	req:	request object - an access to the element storage
		//	type:	kind of the gradient: stroke_painter or fill_painter?
		//Returns NULL, when there is no requested gradient.
		static painter * from(styleset &style, request &req, attribute::kind type)
		{
			painter *result = NULL;

			if(style.has(type)) 
			{
				result = (painter *) req.get_element(style.of(type).val.id,"painter");
			}

			return result;
		}
	protected:
		//////////////////////////////////////////////////////////////////////
		// New virtual methods
		//////////////////////////////////////////////////////////////////////
		//Copies settings from another gradient. Called within initialize() method 
		//Arguments:
		//	req:	a request object (request::initialize)
		//	src:	another painter - source of the information (of the same class!)
		//This method exists, because in SVG one gradient/pattern element can 
		//reference another, and the sum of the attributes from the BOTH gives 
		//the all information required to create a gradient/pattern
		virtual void use(request &req, painter *src)
		{
			if(m_relative_units < 0 && src->relative_units() >= 0) //initialized? 
			{
				m_relative_units = src->relative_units();
			}

			if(m_relative_content_units < 0 && src->relative_content_units() >= 0) //initialized? 
			{
				m_relative_content_units = src->relative_content_units();
			}

			//pattern/gradientTransform:
			//These transfroms are already inverted:
			if (is_empty(src->extra_transform()))
			{
				if(is_empty(m_extra_trans)) 
				{
					m_extra_trans = transform(); //identity
				}
			}
			else //src has valid transformation:
			{
				if(is_empty(m_extra_trans)) //not set in this pattern:
				{
					m_extra_trans = src->extra_transform(); 
				}
			}
		}

		//Initializes the gradient/pattern, when the object receives a 
		//reqest::initialize
		//Arguments:
		//	req:	req::initialize - passed to give the access to the element and
		//			attribute storage.	
		virtual void initialize(request &req)
		{
			if (initialized()) return; //nothing to do
			painter *reference = (painter *) req.get_element(m_href);
			if (reference) 
			{
				//give it a chance to initialze itself, before it 
				//initializes the others (it sometimes may happen, that 
				//the object referenced is after the reference user!):
				if (!reference->initialized()) reference->on_request(req);
				//OK, initialize it!
				use(req, reference);
			}
			
			if (m_relative_units < 0) 
			{					      //default:	
				m_relative_units = 1; //gradient/patternUnits = "objectBoundingBox"
			}

			if (m_relative_content_units < 0) 
			{								  //default:	
				m_relative_content_units = 0; //patternContentUnits = "userSpaceOnUse"
			}
			
			//Give the chance to the children elements to initalize
			element::forward(req);

			m_initialized  = true;
		}

		//////////////////////////////////////////////////////////////////////
		// Other helper methods
		//////////////////////////////////////////////////////////////////////
		//Helper function: changes percent coordinates into simple fractions
		//(it is OK for this specific case - gradients)
		//Arguments:
		//	prs:	the parser
		//	value:	text, that will be interpreted as coordinate (may be 
		//			relative, in %)
		static double parse_coordinate(parser &prs, const char *value)
		{
			bool percent = false;
			return prs.to_length(value, &percent);
		}
		
	protected:
		//////////////////////////////////////////////////////////////////////
		// Instance additional fields
		//////////////////////////////////////////////////////////////////////


		transform	m_internal_trans;//the transfromation based on the gradient/
									 //pattern coordinates. (for gradients: (x1,y1, 
									 //x2,y2) or (cx,cy,r,fx,fy))
									 //reduces gradients to a horizontal span of 
									 //size/radius = 1.0, and center at 0.0,0.0.
									 //for patterns - reduces them to a rectangle
									 //of width = 1.0, and proportional height,
									 //placed at 0,0 (upper - left corner)

		transform	m_unit_trans;	 //for userSpaceOnUse - an identity matrix
									 //for objectBoundingBox - transformation
									 //from BoundingBox of the object into an unit
									 //vector. (Bounding box of the object => an
									 //unit square). In these units are expressed 
									 //the gradients coordinates

		transform	m_extra_trans;	 //the value of additional, explicit attribute
									 //of patter/gradientTransform (additional transform
									 //to object c

		transform	m_actual_trans; //actual reversed transformation (from
									//global cooridnate system to the gradient/pattern
									//coordinate system - a system, in which
									//the gradient_function class calculates 
									//the gradient value, based on received 
									//parameters). Can be an unit transform
									//It is a result of inversed global matrix and
									//all gradient/pattern transfromation matrices.
		
		//gradientUnits / patternUnits:
		int	m_relative_units;//tristate: -1, when not initialized, 
								 //0: when userSpaceOnUse is declared
								 //1: when the objectBoundingBox is declared

		//patternContentUnits:
		int	m_relative_content_units; //tristate: -1, when not initialized, 
								 //0: when userSpaceOnUse is declared
								 //1: when the objectBoundingBox is declared

		handle		m_href;	   //optional: reference to another gradient definition
							   //which attributes can override these ones
							   //can be invalid_handle

		bool		m_initialized; //a flag, that allows to check if the initialized()
								   //on this object was called at least once.

}; //painter

}; //namespace svg
#endif SVG_PAINTER_INCLUDED
