//----------------------------------------------------------------------------
// SVG Viewer - Version 1.0
// A library of SVG renderer
// Copyright (C) 2008 Witold Jaworski
// Contact:	witold.jaworski@airplanes3d.net
//			http://airplanes3d.net
// 
// Contains elements from:
//			AGG - a graphic library (Copyright Maxim Shemanarev, mcseem@antigrain.com)
//          http://antigrain.com
// Uses functions from:
//          Expat - a XML parser library (Copyright James Clark)
// 
// SVG Viewer is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
// 
// SVG Viewer is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with SVG Viewer; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, 
// MA 02110-1301, USA.
//----------------------------------------------------------------------------

#ifndef SVG_PARSER_INCLUDED
#define SVG_PARSER_INCLUDED
#include "svg_basics.h"
#include "agg_svg_path_tokenizer.h"

//Maximum length of stored identifiers, including trailing null
#define SVG_MAX_NAME 64 
#define SVG_MAX_ATTRIB_NAME 64	
//does not apply to the "style" and "d" attributes - they can be longer! 
#define SVG_MAX_ATTRIB_VALUE 256
									

typedef agg::svg::path_tokenizer  path_tokenizer;

namespace svg
{
//////////////////////////////////////////////////////////////////////
// reference: forward delclaration
//////////////////////////////////////////////////////////////////////
#ifndef SVG_ELEMENT_INCLUDED
class	element;
#endif
//////////////////////////////////////////////////////////////////////
// Helper structure - a placeholder for the element, which can be not 
// loaded (yet). It is kept in the root's element_storage vector.
//////////////////////////////////////////////////////////////////////
struct reference 
{
	char	name[SVG_MAX_NAME]; //xml id of this element
	bool	presisted;			//true, if this instance is presisted with the
								//new operator (if true, the elm pointer will 
								//be released in the destructor)
	element * elm;				//pointer to the element (NULL for unitialized)
	//Beware: the pointer stored in elm MUST me created with the new operator!
	//(see code of parser::start_element() function)

	//empty constructor
	reference() {name[0] = 0; elm = NULL; presisted = false;}
	
	//copy constructor - just to control the unsafe copying its
	//precious content - a heap-allocated elem!
	reference(reference &r);

	//Initializes the reference WITHOUT any existing element
	//Arguments:
	//	ename:	name of the element (the xml id attribute value)
	//Beware: it is assumed, that it will receive the assign() call, later
	reference(const char *ename) 
	{//sometimes the requested name can be longer: in such case - ignore it!
		if (strlen(ename) < sizeof(name)) 
		{
			strcpy(name, ename);
		}
		else
		{
			name[0] = 0;
		}
		elm = NULL; 
		presisted = false;
	}

	//assigns an element to this reference:
	void assign(element *e) {elm = e; presisted = (e != NULL);}

	//Beware: elm MUST me created with the new operator!
	//(see code of parser::start_element() function)
	~reference();
};

//////////////////////////////////////////////////////////////////////
// SVG parser
// used by SVG graph root node to load the SVG file.
//////////////////////////////////////////////////////////////////////
class parser  
{
friend class element;
    public:
		//Constructor
		//Arguments:
		//	paths:	path storage, that will be filled with element's paths
		//	attrs:	attribute storage, that will be filled with element's attributes
		//	elems:	element storage, that will be filled with the SVG graph elements
		//			(instances are created on the heap, using new operator)
		parser(path_storage &paths, attr_storage &attrs, element_storage &elems); 
        ~parser();

		//main method - parse the given file name
		void parse(const char *fname);

        // XML event handlers, required by the Expat - called from the parse() method
		// the data pointer is the pointer to actual parser object
        static void start_element(void* data, const char* el, const char** attr);
        static void end_element(void* data, const char* el);
        static void content(void* data, const char* s, int len);

		//helper method: parse attributes of given SVG element 
		//(to be called inside the element's constructor)
		//Arguments:
		//	elm:	SVG graph element, that is the owner of these attributes
		//	attr:	attributes - even elements: names, odd - values (text)
        void parse(element &elm, const char** attr);

		//////////////////////////////////////////////////////////////////////
		// Element's attributes handling 
		//////////////////////////////////////////////////////////////////////
		//Stores the attribute in the storage (maintained by the root)
		//Arguments:
		//	elm:	the SVG graph element, that attribute belongs to
		//	attr:	an attribute instance (for example - a local variable)
		//Updates the internal state of the object elm - initializes (if needed)
		//m_attribute_first and updates m_attribute_last fields
		void store(element &elm, attribute attr);
		
		//Tries to recognize the name and value of an attribute, as one of the
		//style attributes, and - if it is a style - stores it in the 
		//storage (maintained by the root)
		//Arguments:
		//	elm:	the SVG graph element, that attribute belongs to
		//	name:	name of an attribute
		//	value:	value of an attribute
		//Returns true when the attribute was recognized as a style attribute.
		//Updates the elm intenral stats. (Uses internally the store(elm,attr) 
		//procedure). It is a kind of "default interpreter" of various
		//style attributes.
		bool store_as_style(element &elm, const char * name, const char *value);

		//Returns the attribute, associated with a given handle
		//Arguments:
		//	id:		a handle to an attribute 
		attribute *get_attribute(handle id) 
			{ if (id < m_attrs.size()) return &(m_attrs[id]); else return NULL;}

		//////////////////////////////////////////////////////////////////////
		// Element's presistence handling
		//////////////////////////////////////////////////////////////////////
		//Registers element under given id (name) from the xml file 
		//Arguments:
		//	elm:		element to be registered;
		//	name:		unique name (usually - the xml id attribute) of the elm
		//	shared:		optional: true, when the element <elm> can be referenced by 
		//				other objects
		//  from_xml:	optional: true, when this is from the xml file, (i.e. 
		//				there will be a call to parser::end_element(). When elm 
		//				is an object, created just from the scratch, set it to false

		//Updates the fields m_id and initializes m_id_last in the elm object
		void store(element &elm, const char *name, 
										bool shared = false, bool from_xml = true);

		//Returns an element, assigned to a handle
		//Arguments:
		//	id:		a handle to an element, obtained from get_id()
		element  *get_element(handle id)
			{ if (id < m_elems.size()) return m_elems[id].elm; else return NULL;}

		//Returns an id of element that has given name  (the value of SVG "id" attribute 
		//is called here "name")
		//Arguments:
		//	name:	xml id of the element sought.
		//Beware: it always returns an id, but the requested element can 
		//be not available (retrieve return NULL) until call to initialize()
		//It is not optimized function - it performs linear search, with
		//helper lookup table for the most used elements.
		handle get_id(const char *name);

		//Useful function during reading the content of the container
		//(for the commit() function - to discover the number of child elements)
		//number of the elements in the whole SVG graph. 
		unsigned elements() {return m_elems.size();}

		//////////////////////////////////////////////////////////////////////
		// Specialized, helper functions for the element's attributes:
		//////////////////////////////////////////////////////////////////////

	    //Converts a character string to double
		//Arguments:
		//	str:	the string to be converted.
		//Returns: just a double
		static double to_double(const char* str);

		//Converts a character string to a valid color value
		//Arguments:
		//	str:	the string to be converted
		//Returns: an RGBA color
		static agg::rgba8 to_color(const char* str);

		//Converts a character string to a valid length
		//Arguments:
		//	str:		string, which can look like "14.5cm" or "14.5" or "14.5%"
		//	relative:	optional: if not null, set to true, when a percent value
		//				was specified. Otherwise - set to false.
		//				It is up to the caller, to do something with the relative 
		//				value.
		//Returns: length in pixels (i.e. "user coordinates" in the SVG spec), or
		//for the relative values, the value as simple fractal.
	    static double to_length(const char* str, bool *relative = NULL);

		//Helper function - for various string tokenization
		//Arguments:
		//	text:		source text;
		//	sep:		separator character
		//	buf:		a buffer for the result (first found token)
		//	bufsize:	size of the buffer (including trailing null)
		//Returns: pointer to a first character after the separator, or 
		//the trailing end of the source text, when no more separators were found
		//Fills the buffer with the token found, stripped of whitespaces.
		//This function can be used in many cases of the string iteration
		static char *get_first_token(const char *text, char sep, char *buf, 
																	size_t bufsize);
		//    * You are always missing yet another string function ! *
		//helper function, that removes the whitespaces from the source
		//stores the result in another buffer.
		//Arguments:
		//	srcbuf:		source string;
		//	dstbuf:		destination string
		//	max_size:	maximum number of characters that can be copied into dst
		//				(without trailing null)
		//Returns:  buf, which contains the copied text
		//			NULL, when the src is longer than max_size 
		char *remove_spaces(const char *srcbuf, char *dstbuf, size_t max_size);

		//parse the "d" attribute - path coordinates and stores it into a path_storage
		//Arguments:
		//	str:	the value of the svg::d xml attribute - geometry of the path		
		//the handle, returned by this function, is a handle to a path, not to an attribute!
        handle parse_d(const char* str);
		
		//Helper function - creates a single line segment 
		//Arguments:
		//	x1,y1:	coordinates of the begining
		//	x2,y2:	coordinates of the end
		//the handle, returned by this function, is a handle to a path, not to an attribute!
	    handle create_line(double x1, double y1, double x2, double y2);

		//Helper function - creates an ellipse
		//Arguments:
		//	cx,cy:	center of the ellipse
		//	rx,ry:	elipse's radii
		//the handle, returned by this function, is a handle to a path, not to an attribute!
		handle create_ellipse(double cx, double cy, double rx, double ry);

		//Helper function - creates a rectangle (optional rounded corners)
		//Arguments:
		//	x,y:	left, top corner of the rectangle (smallest x, y coordinates)
		//	w,h:	width and height of the rectangle
		//	rx,ry:	corner's radii (can be 0)
		//the handle, returned by this function, is a handle to a path, not to an attribute!
		handle create_rect(double x, double y, double w, double h, double rx, double ry);

		//parse the "points" attribute - polygon/polyline coordinates and stores it into a path_storage
		//Arguments:
		//	str:		the value of the svg::points xml attribute - vertices of the polygon/polyline		
		//	close_flag:	true, if the element is svg:polygon, and false for the svg:polyline		
		//the handle, returned by this function, is a handle to a path, not to an attribute!
        handle parse_points(const char* str, bool close_flag);
		
		//parses the content of a "fill" or "stroke" attribute:
		//		a color
		//	or
		//	-	an url to gradient or pattern
		//Arguments:
		//	elm:	the SVG graph element, that attribute belongs to
		//	type:	fill_flag or stroke_flag
		//	value:	value of an attribute (sometimes - a specification, prepared before)
		//Returns true when the attribute was recognized and stored.
		//Updates the elm intenral stats. (Uses internally the store(elm,attr) 
		//procedure). 
		bool parse_painter(element &elm, attribute::kind type, const char *value);

		//default handler for parsing the style expression
		//Arguments: 
		//	elm:	element, which contains this style attribute
		//	str:	the value of the xml svg::style attribute
		//It recursively calls the elm.on_attribute() for every attribute in the style string
        void parse_style(element &elm, const char* str);

		//default handler for parsing the transformation expression
		//Arguments: 
		//	str:	the value of the xml svg::transform attribute
		//It returns a transformation structure
        transform parse_transform(const char* str);
		
	private:        
		//Assigns the element to the existing handle
		//Updates internal state of the element
		//Arguments:
		//	id:		the valid (existing!) handle id
		//	elm:	element, to be assigned
		//	cref:	optional: a cross - reference to the additional "use" placeholder
		void assign(handle id, element *elm, handle cref = invalid_handle);

		//helper function for parsing a single style "name:value" pair
        void parse_style_element(element &elm, const char* nv_start,const char* nv_end);

		//helper function to accumulate some attributes, that have to be
		//accumulated before final sythesis into a stored attribute.
		//(An example of such case are attributes that build the font definition:
		// the font-family, font-style, font-size)
		//Arguments:
		//	name:	attribute's name
		//	value:	attribute's value
		//Function returns true, when the attribute was accumulated (i.e. 
		//it should not be passe to the element in that from)
        bool attribute_accumulated(const char* name,const char* value);

		//helper function that returns a handle to an object, that matches the 
		//requested specification. Used for the dasharrays, fonts, etc.
		handle get_id(attribute::kind type, element &parent, const char *spec);
        
    private:

		path_storage				&m_paths;
		attr_storage				&m_attrs;
		element_storage				&m_elems;
		agg::pod_bvector<handle>	m_lookup;//helper - lookup table for the 
											 //referenced items
		element						*m_element;	//last element stored in the m_elems
		unsigned					m_level;	//actual nest level in the xml source
		agg::svg::path_tokenizer	m_tok;		//helper object, used to parse "d" attributes

		//buffers used to parse style - in parse_style() function
		//(I have decided to use a static buffers, because whole parser instances
		//will be created on a stack of the root constructor - and dissapear silenly
		//after it, not allocating anything important on the heap)
		//buffer for a style attribute name 
		char				m_style_name[SVG_MAX_ATTRIB_NAME]; 
		//buffer for a style attribute value (not for "d" nor "style"):
		char				m_style_value[SVG_MAX_ATTRIB_VALUE];

		//helper specialistic buffers for font parsing:
		char	m_font_family[SVG_MAX_NAME]; //according Win32 spec, 32 chars per face
		char	m_font_style[SVG_MAX_NAME/2]; //just "normal" | "italic" | "oblique"	
		char	m_font_weight[SVG_MAX_NAME/2]; //just "normal" | "bold" | "nnn"
		
}; //parser


}; //namespace svg
#endif SVG_PARSER_INCLUDED
