//----------------------------------------------------------------------------
// SVG Viewer - Version 1.0
// A library of SVG renderer
// Copyright (C) 2008 Witold Jaworski
// Contact:	witold.jaworski@airplanes3d.net
//			http://airplanes3d.net
// 
// Contains elements from:
//			AGG - a graphic library (Copyright Maxim Shemanarev, mcseem@antigrain.com)
//          http://antigrain.com
// Uses functions from:
//          Expat - a XML parser library (Copyright James Clark)
// 
// SVG Viewer is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
// 
// SVG Viewer is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with SVG Viewer; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, 
// MA 02110-1301, USA.
//----------------------------------------------------------------------------

#ifndef SVG_PATH_TEXT_POSITION_INCLUDED
#include "svg_text_position.h"
#include "svg_conv_locator.h"
#define SVG_PATH_TEXT_POSITION_INCLUDED

namespace svg
{
//////////////////////////////////////////////////////////////////////
// linear_text_position: it is a helper object, used to draw text strings,
// which are not placed on a path (see svg::text element)	
//////////////////////////////////////////////////////////////////////
class path_text_position:public text_position  
{
public:
	//constructor: creates a text position in an uninitioalized state
	//Constructor
	//Arguments:
	//	vs:			source of the
	//	path_id:	id of the path from the vs
	//	position:	initial position on the path (negative, when relative (%)!)
	//	trans:		global transformation, that has to be applied to vs
	path_text_position(vsource &vs, handle path_id, double position, 
												const transform &trans): 
	text_position(),
	m_path(vs,path_id,trans)
	{
		if (position < 0) //when negative: relative! (0.0 .. 1.0)
		{
			m_t = -position * m_path.length();
		}
		else
		{
			m_t = position; //initial position: 0
		}//end if

	}

	//Sets the initial current text position
	//(the y parameter is ignored)
	virtual void set(double t, double ignored, string &text, unsigned pos, 
					unsigned count, double dx, double dy, double rotation)
	{
		text_position::set(ignored,ignored,text,pos,count,dx,dy,rotation);

		if(m_wmode == tb) m_wmode = lr; //ignore top-bottom directions!

		if (is_valid())
		{
			rectangle len = offset(); //length of the text

			m_t = t - (m_anchor * len.x2);

			//text position is valid, when it fits into curve
			if (m_t < 0.0) 
			{
				m_valid = false;
			}
			else
			{
				m_valid = m_path.go_to(m_t); 
			}//end if
		}//end if
	}
	

	//Current text position on the path:
	virtual double x() 
	{ 
		return m_t;
	}; 

	virtual double y() 	
	{
		return 0;
	}; 

	//actual glyph transformation
	virtual transform transformation()
	{
		transform t = internal_transform();
		t *= m_local;
		//coordinates on the path:
		double xs,ys, xm, ym, xe, ye;	//coordinates: <s>start, <m>iddle, <e>nd
		double span = advance_x()*0.5;	//half of the text size
		rectangle box = m_glyph->bounds();

		m_path.position(&xs, &ys); //start point
			//for first visible character from a text, in which first part has
			//fallen apart the path, the -span <= m_t <= span, and position() = 0;
			m_valid &= m_path.move(m_t - m_path.position() + span); 
		m_path.position(&xm, &ym); //mid point
			m_valid &= m_path.move(span);
		m_path.position(&xe, &ye); //end point

		//transform coords into more "handy" segments:
		segment base(xs, ys, xe, ye); //text advance direction

		//origin of the glyph: back from midpoint, by half of the charwidth (-span):
		segment origin(xm, ym, - span, base.dx(), base.dy());
		
		t.rotate(base.angle()); 
		t.translate(origin.x(), origin.y());

		return t;
	};

	//advances text position to the next point 
	//(and to next character within text chunk)
	virtual bool advance()
	{
		if (m_pos < m_count)
		{
			m_t += advance_x();
			
			m_pos++;

			if (m_glyph) m_glyph->select(code(m_pos));
			//Inkscape does not take the kerning into 
			//count during advanced the character:
			if (!m_auto_kerning) //Suprising, but for the auto-kerninig
			{					 //it is not applied to the distance between
								 //letters!
				m_t += kerning_x()*m_scale;
			}
			
			//m_t is determined, let's move:
			if (m_t < m_path.position()) 
			{
				if (m_t < -0.5*advance_x()) //We are starting to draw the letters
				{							//when their midpoin falls on the path
					m_valid = false;
				}
				else //go_to() returns false, when m_t falls behind the end
				{	 //of the path:
					m_valid = m_glyph && m_path.go_to(m_t);
				}
			}
			else
			{
				m_valid = m_glyph && m_path.move(m_t - m_path.position());
				if (m_valid) //does entrie character fit on the path?
				{
					m_valid = (m_path.position() + 0.5*advance_x()) <= m_path.length();
				}
			}

		}
		return (m_pos < m_count);
	}; 
	
protected:
		conv_locator	m_path;	//the curve (path)
		double			m_t;		//actual position on curve
}; //path_text_position

}; //namespace svg
#endif SVG_PATH_POSITION_INCLUDED
