//----------------------------------------------------------------------------
// SVG Viewer - Version 1.0
// A library of SVG renderer
// Copyright (C) 2008 Witold Jaworski
// Contact:	witold.jaworski@airplanes3d.net
//			http://airplanes3d.net
// 
// Contains elements from:
//			AGG - a graphic library (Copyright Maxim Shemanarev, mcseem@antigrain.com)
//          http://antigrain.com
// Uses functions from:
//          Expat - a XML parser library (Copyright James Clark)
// 
// SVG Viewer is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
// 
// SVG Viewer is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with SVG Viewer; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, 
// MA 02110-1301, USA.
//----------------------------------------------------------------------------

#ifndef SVG_RADIAL_GRADIENT_INCLUDED
#include "svg_gradient.h"
#define SVG_RADIAL_GRADIENT_INCLUDED

namespace svg
{

//----------------------------- Offset -------------------------------
//////////////////////////////////////////////////////////////////////
// radial_gradient: one of the two gradient types, used in SVG. 
//			        (the second one is the linear_gradient)
//////////////////////////////////////////////////////////////////////
class radial_gradient : public gradient 
{
	public:
		//////////////////////////////////////////////////////////////////////
		// constructor and overriden virtual methods
		//////////////////////////////////////////////////////////////////////

		radial_gradient(element *parent, unsigned level): gradient(parent, level),
		m_cx(0.5), m_cy(0.5), 
		m_fx_valid(false), m_fx(m_cx), m_fy_valid(false), m_fy(m_cy), 
		m_r(0.0)
		{}

		virtual const char * tag() {return "radial_gradient";};

		virtual void on_attribute(parser &prs, const char * name, const char *value)
		{
			if (strcmp(name, "cx") == 0) 
			{	
				m_cx = parse_coordinate(prs,value);
			}
			else 
			if (strcmp(name, "cy") == 0) 
			{	
				m_cy = parse_coordinate(prs,value);
			}
			else 
			if (strcmp(name, "fx") == 0) 
			{	
				m_fx_valid = true;
				m_fx = parse_coordinate(prs,value);
			}
			else 
			if (strcmp(name, "fy") == 0) 
			{	
				m_fy_valid = true;
				m_fy = parse_coordinate(prs,value);
			}
			else
			if (strcmp(name, "r") == 0) 
			{	
				m_r = parse_coordinate(prs,value);
			}
			else 
			{//if the style is not recognized:
				gradient::on_attribute(prs,name,value);
			}
		}
		
		virtual bool is_of(const char *cls) 
		{
			if(strcmp("radial_gradient",cls) == 0)
			{
				return true;
			}
			else //it is not your name? Make sure, asking your parent :)
				return gradient::is_of(cls);
		}

	protected:
		//Initializes the gradient, when the object receives a reqest::initialize
		//Arguments:
		//	req:	req::initialize - passed to give the access to the element and
		//			attribute storage.	
		//
		// BEWARE: in the child classes initialize the m_span, before calling this one! 
		//
		virtual void initialize(request &req)
		{

			//in further processing, m_fx is relative to the center - cx, cy
			if (m_fx_valid) m_fx -= m_cx; else m_fx = 0.0;
			if (m_fy_valid) m_fy -= m_cy; else m_fy = 0.0;
			//OK, I do not know why, but this experimental coefficent = 0.6
			//makes the radial gradient effect properly matched the effect
			//that is visible in the Inkscape. Is it possible, that AGG uses
			//slightly different mathematical model of the focus!?
			m_cy += 0.6*m_fy;
			m_cx += 0.6*m_fx;

			m_span = segment(m_cx, m_cy, m_cx+m_r, m_cy); //since this early moment.
													  //do not in this class the
													  //m_cx, m_cy, m_r -
													  //use m_span instead!
			gradient::initialize(req);
		}

		//Returns an offset generator object matching the given symbol
		//Arguments:
		//	symbol:	one of three values: "repeat", "reflect", "pad"
		//Each child class should return a correspodnig offset generator
		virtual svg::offset_generator *select_generator(const char *symbol)
		{
			if(strcmp(symbol, "reflect") == 0)
			{
				m_src_reflect.initialize(m_r, m_fx, m_fy);
				return &m_src_reflect;
			}
			else
			if(strcmp(symbol, "repeat") == 0)
			{
				m_src_repeat.initialize(m_r, m_fx, m_fy);
				return &m_src_repeat;
			}
			else //default: pad
			{
				m_src_pad.initialize(m_r, m_fx, m_fy); 
				return &m_src_pad;
			}
		}


	protected:
		template<class GradientAdaptor> 
		class offset_source : public svg::offset_generator
		{
			public:
			offset_source(): m_adaptor(m_function) {}

			void initialize(double r, double fx, double fy)
			{
				m_function.init(r, fx, fy);
			}

			virtual int calculate(int x, int y, int d) const
			{
				return m_adaptor.calculate(x, y, d);
			}

			protected:
			agg::gradient_radial_focus m_function;
			GradientAdaptor m_adaptor;
		};//end offset_source

		typedef agg::gradient_reflect_adaptor<agg::gradient_radial_focus>	
																reflect_adaptor;
		typedef agg::gradient_repeat_adaptor<agg::gradient_radial_focus>	
																repeat_adaptor;
		typedef gradient_clamp_adaptor<agg::gradient_radial_focus>			
																pad_adaptor;

		//////////////////////////////////////////////////////////////////////
		// Instance additional fields
		//////////////////////////////////////////////////////////////////////
		//Gradient coordinates
		double	m_cx, m_cy;		//the center of the circle
		double	m_r;	//radius

		bool	m_fx_valid, m_fy_valid;	//false, when focus is not valid (was not set)
		double	m_fx, m_fy;		//focus point

		//Alternative offset sources:
		 offset_source<reflect_adaptor>	m_src_reflect;
		 offset_source<repeat_adaptor>	m_src_repeat;
		 offset_source<pad_adaptor>		m_src_pad;
}; //radial_gradient

}; //namespace svg
#endif SVG_RADIAL_GRADIENT_INCLUDED
