//----------------------------------------------------------------------------
// SVG Viewer - Version 1.0
// A library of SVG renderer
// Copyright (C) 2008 Witold Jaworski
// Contact:	witold.jaworski@airplanes3d.net
//			http://airplanes3d.net
// 
// Contains elements from:
//			AGG - a graphic library (Copyright Maxim Shemanarev, mcseem@antigrain.com)
//          http://antigrain.com
// Uses functions from:
//          Expat - a XML parser library (Copyright James Clark)
// 
// SVG Viewer is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
// 
// SVG Viewer is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with SVG Viewer; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, 
// MA 02110-1301, USA.
//----------------------------------------------------------------------------

#ifndef SVG_REQ_FILTER_INCLUDED
#include "svg_basics.h"
#include "svg_segment.h"
#include "svg_parser.h"
#include "svg_request.h"
#include "svg_shape.h"
#define SVG_REQ_FILTER_INCLUDED
namespace svg
{
//////////////////////////////////////////////////////////////////////
// reference: forward delclaration
//////////////////////////////////////////////////////////////////////
#ifndef SVG_FILTER_INCLUDED
class	filter;
#endif

//////////////////////////////////////////////////////////////////////
// SVG req_filter: this message is send to a very specific element group:
// the filter primitives. It serves them the input and output buffers
// (in the form of the pixel formats).
// Before use, send it to the parent in a req_prepare, with request::prepare
// to allow it assign the element and attribute storage.
//////////////////////////////////////////////////////////////////////
struct req_filter: public request
{

	public:
		//Constructor
		//The svg::root element should assign the SVG graph attrrs and elems
		//storages! Then it will come back to the sender.
		//Arguments:
		//	fltr:	the filter definition
		//	elm:	element, wich has to have this filter applied
		//	gtrans:	global transformation
		//	bounds:	actual clip box
		//	pixels:	traget pixel format (result buffer)
		//	extent_ratio: optional. determines, how bigger will be the intermediate
		//				  buffers comparing to the target area (some extent is required
		//				  for example for the gaussian blurs - otherwise they would be 
		//				  distored at the edges). Default: 2.0 (50% from each side)
		req_filter(svg::filter &fltr,shape &elm, transform gtrans, 
												rectangle bounds, pixfmt &pixels, 
														 double extent_ratio = 2.0);

		inline unsigned x() {return m_x;}
		inline unsigned y() {return m_y;}
		inline unsigned width() {return m_width;}
		inline unsigned height() {return m_height;}

		//false, when filter is empty: 
		inline bool is_valid() {return m_width && m_height;}

		//Method translates the nominal coordinates of filter primitives 
		//into pixel coordinates of the result buffer (i.e. the buffer
		//available through get("") call). 
		//Arguments::
		//	prm_x:	nominal x coordinate, as specified in the XML source
		//	prm_y:	nominal y coordinate, as specified in the XML soirce
		//	px:		address of the result x cooridnate
		//	py:		address of the result y coordinate
		void prm_to_pixels(double prm_x, double prm_y, int *px, int *py)
		{
			m_prm_trans.transform(&prm_x, &prm_y);
			*px = agg::iround(prm_x);
			*py = agg::iround(prm_y);
		}

		//For a length, expressed in the nominal coordinates of filter primitves
		//this function returns corresponding length in the pixel coordinates of
		//of the result buffer (i.e. the buffer available through get("") call)
		//Arguments:
		//	len:		nominal length, as specified in the XML source
		//	share_x:	share of the scale in the x direction
		//				(scaling in x and y direction can not be uniform)
		unsigned prm_to_length(double len, double share_x = 0.5)
		{
			if (share_x < 0.0) share_x = 0.0;
			if (share_x > 1.0) share_x = 1.0;
			double x0 = 0.0, y0 = 0.0;
			double x1 = len, y1 = 0.0;
			double x2 = 0.0, y2 = len;
			m_prm_trans.transform(&x0,&y0);
			m_prm_trans.transform(&x1,&y1);
			m_prm_trans.transform(&x2,&y2);
			segment tx(x0,y0,x1,y1);
			segment ty(x0,y0,x2,y2);
			return agg::uround(tx.length() * share_x + ty.length() * (1.0 - share_x));
		}
	public:
		///////////////////////////////////////////////////////////////////
		// Base class for the filter buffers (SourceGraphics, BacgroundAlpha,...)
		///////////////////////////////////////////////////////////////////
		class buffer
		{
			public:
				buffer(): m_pixels(m_rbuf) {}
				virtual ~buffer() {} //useful ti release the intermediate buffers,
									 //allocated on the heap (see plain_buffer)

				//returns the content of the buffer:
				inline pixfmt_pre &pixels() {return m_pixels;}
				inline rendering_buffer &rbuf() {return m_rbuf;}

			protected:
				rendering_buffer m_rbuf;
				pixfmt_pre		 m_pixels;
		}; //buffer

		//Returns the pixel format having given name. If the name does not
		//exists, yet - it creates it as a plain buffer, and then returns!
		//Arguments:
		//	name:	name of the buffer (local to particular filter element)
		pixfmt_pre &get(const char *name)
		{
			return select(name).buf().pixels();
		}

		//Similar to get(<name>), but returns the rendering buffer.
		//seldom used.
		rendering_buffer &get_buf(const char *name)
		{
			return select(name).buf().rbuf();
		}

	protected:
		///////////////////////////////////////////////////////////////////
		// Helper structure - to keep the buffer references in the pod_bvector 
		///////////////////////////////////////////////////////////////////
		struct reference
		{
			public:
			//Creates a new reference: 
			//Arguments
			//	name:	name of the buffer (can be "" for the default)
			reference():
			m_buf(NULL)
			{
				m_name[0] = m_name[sizeof(m_name)-1] = 0; //prepare trainling nulls
			}
			
			virtual ~reference()
			{
				assign("", NULL);
			}

			//Assigns a picture buffer to this reference
			//(to not allow  an intermediate object to delete this pointer)
			//Arguments:
			//	buf:	a picture buffer - HAS to be allocated on the heap 
			//			by the new operator!
			void assign(const char *name, buffer *buf)
			{
				if (m_buf) delete m_buf;
				m_buf = buf;
				strncpy(m_name,name,sizeof(m_name)-1); //cut the longer names
			}

			//Used in req_filter::get_buf to find the matching buffer
			//(says always "no", when there is no buffer assigned)
			bool is(const char *name) {return m_buf && (strcmp(m_name, name) == 0);}
			
			//returns the content of the reference
			inline buffer &buf() {return *m_buf;}

			protected: 
				char m_name[SVG_MAX_NAME]; //name of the buffer (default is "")
				buffer * m_buf;			  //a pointer to the allocated buffer
	
		}; //reference

	protected:
		//Adds a new intermediate buffer, having given name
		//Arguments:
		//	name:	name of the new picture buffer
		//			Predefined SVG names, as:
		//				"SourceGraphic"
		//				"SourceAlpha"
		//				"BackgroundImage"
		//				"BackgroundAlpha"
		//				"FillPaint"
		//				"StrokePaint"
		//			are recognized and treated according SVG spec
		reference &add(const char *name);

		//Returns the reference structure having given name. If the name 
		//does not exists, yet - it creates it as a plain buffer, and then 
		//returns!
		//Arguments:
		//	name:	name of the buffer (local to particular filter element)
		reference &select(const char *name);

	protected:	
		svg::filter	&m_filter;	//filter, which created this request
		shape		&m_elm;		//Element, which apply the filter
		pixfmt		&m_pixels;	//result buffer
		agg::pod_bvector<reference> m_buffers; //buffers for the source and 
											   //intermediate results

		//filter region, expressed in the pixel units of the m_pixels buffer:
		int	m_x;
		int	m_y;
		int	m_width;
		int	m_height;

		transform	m_trans;//global transformation, which  should be applied
							//to m_elem, to place it in the proper position
							//of the filter region (usually the global trans-
							//formation, translated by m_x, m_y)
		transform	m_prm_trans; //transformation used in prm_to_pixel() method:
								 //form nominal (used by filter primitives) coords
								 //to the result buffer coords.
}; //req_filter

}; //namespace svg
#endif SVG_REQ_FILTER_INCLUDED
