//----------------------------------------------------------------------------
// SVG Viewer - Version 1.0
// A library of SVG renderer
// Copyright (C) 2008 Witold Jaworski
// Contact:	witold.jaworski@airplanes3d.net
//			http://airplanes3d.net
// 
// Contains elements from:
//			AGG - a graphic library (Copyright Maxim Shemanarev, mcseem@antigrain.com)
//          http://antigrain.com
// Uses functions from:
//          Expat - a XML parser library (Copyright James Clark)
// 
// SVG Viewer is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
// 
// SVG Viewer is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with SVG Viewer; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, 
// MA 02110-1301, USA.
//----------------------------------------------------------------------------
#include "svg_exception.h"
#include "svg_painter.h"
#include "svg_req_render.h"
#include "svg_styleset.h"
#include "svg_dash.h"
#include "svg_earray.h"
#include "svg_path_text_position.h"
#include "agg_bounding_rect.h"

namespace svg
{
	//Measures an element, described by the vertices at given path and 
	//local transformation 
	//Arguments:
	//	path_id:	id of a path, stored in the m_paths
	//	ltrans:		local transformation of a path - relative to
	//	lstyle:		style, that has to be used.
	void req_measure::perform(handle path_id, transform &ltrans, styleset &lstyle)
	{
		rectangle pb(0,0,-1,-1); //for the bounds of this path (having this <path_id>)
		m_trans = ltrans;
		m_trans *= trans; //m_trans: effective transformation
		m_curved_count.count(0); //reset the counter of local vertices
		if (agg::bounding_rect_single(m_curved_count, path_id, 
											&(pb.x1), &(pb.y1), &(pb.x2), &(pb.y2)))
		{
			if (bounds.is_valid())
			{
				bounds = agg::unite_rectangles(bounds,pb);
			}
			else //special case: first shape - nothing to compare:
			{
				bounds = pb;
			}

			vertex_count += m_curved_count.count();
		}

	}

//----------------------------------------------------------------------------
	//Standard way of rendering a path, stored in the path_storage:
	//Arguments:
	//	path_id:	the id of the path, stored in the path_storage, to be rendered
	//	ltrans:		local transformation matrix
	//	lstyle:		style, that has to be used.
	void req_render::perform(handle path_id, transform &ltrans, styleset &lstyle)
	{
		double scale;
		if(has(no_stroke)) //clear the 
		{
			lstyle.to_plain_contour(); //remember: in svg::path, svg::tspan, and  
									   //all other objects that deliver the styleset
									   //to this method always use a local style
									   //structure, on the stack of shape.render() method!
									   //(because it can be modified here)
		}
		
		//m_trans has to have ultimate transfromation:
		m_trans = ltrans; //it is built from the local transfrom (within the group)...
		m_trans *= trans; //... and the group - actual "global" transformation

		scale = m_trans.scale();
        m_curve.approximation_scale(scale);
        m_curve.angle_tolerance(0.0);

        if(lstyle.of(attribute::fill_flag).val.boolean)
        {
 	        m_ras.reset();
			if (lstyle.has(attribute::fill_rule))
			{
				m_ras.filling_rule(lstyle.of(attribute::fill_rule).val.filling_rule);
			}

            if(fabs(m_contour.width()) < 0.0001)
            {
				m_ras.add_path(m_c_curve, path_id);
            }
            else
            {
				m_contour.miter_limit(lstyle.of(attribute::stroke_miterlimit).val2double());
				m_ras.add_path(m_contour, path_id);
            }

			switch(lstyle.of(attribute::fill_mode).val.draw_mode)
			{
			case solid:
				lstyle.apply_fill_color(m_rns);
	            agg::render_scanlines(m_ras, m_scl, m_rns);
				break;

			case other:
				svg::painter *paint = painter::from(lstyle,*this,attribute::fill_painter);
				if (paint && paint->initialized())
				{
					paint->set(painter_trans ? *painter_trans : m_trans, lstyle.fill_opacity());
					agg::render_scanlines_aa(m_ras, m_scl, m_rnb, m_allocator, *paint);
				}
				break;
			}
        }

        if(lstyle.of(attribute::stroke_flag).val.boolean)
        {
			//This element is null, when there is no dash settngs
			//(because its style contains the val.id = invalid_handle)
			element *dash = get_element(lstyle.of(attribute::stroke_dasharray).val.id);
						
			if (dash)
			{
				((svg::dash *) dash)->apply(*this,m_dash); //applies the dash pattern
				m_dash.dash_start(lstyle.of(attribute::stroke_dashoffset).val.real);

				lstyle.apply_stroke(m_dash_stroke); //applies the width and rounding
				m_dash_stroke.approximation_scale(scale);
			}
			else 
			{
				lstyle.apply_stroke(m_solid_stroke); //applies the width and rounding
				m_solid_stroke.approximation_scale(scale);
			}
			// If the *visual* line width is considerable we 
			// turn on processing of curve cusps.
			//---------------------
			if(lstyle.of(attribute::stroke_width).val.real * scale > 1.0)
			{
				m_curve.angle_tolerance(0.2);
			}

			m_ras.reset();
			m_ras.filling_rule(agg::fill_non_zero);

			if (dash)
			{
				m_ras.add_path(m_dash_outline, path_id);
			}
			else 
			{
				m_ras.add_path(m_solid_outline, path_id);
			}
			
			
			switch(lstyle.of(attribute::stroke_mode).val.draw_mode)
			{
			case solid:
				lstyle.apply_stroke_color(m_rns);
				agg::render_scanlines(m_ras, m_scl, m_rns);
				break;

			case other:
				svg::painter *paint = painter::from(lstyle,*this,attribute::stroke_painter);
				if (paint && paint->initialized())
				{
					paint->set(painter_trans ? *painter_trans : m_trans,lstyle.stroke_opacity());
					agg::render_scanlines_aa(m_ras, m_scl, m_rnb, m_allocator, *paint);
				}
				break;
			}
        }


	}
//----------------------------------------------------------------------------


} // namespace svg