//----------------------------------------------------------------------------
// SVG Viewer - Version 1.0
// A library of SVG renderer
// Copyright (C) 2008 Witold Jaworski
// Contact:	witold.jaworski@airplanes3d.net
//			http://airplanes3d.net
// 
// Contains elements from:
//			AGG - a graphic library (Copyright Maxim Shemanarev, mcseem@antigrain.com)
//          http://antigrain.com
// Uses functions from:
//          Expat - a XML parser library (Copyright James Clark)
// 
// SVG Viewer is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
// 
// SVG Viewer is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with SVG Viewer; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, 
// MA 02110-1301, USA.
//----------------------------------------------------------------------------

#ifndef SVG_REQUEST_INCLUDED
#define SVG_REQUEST_INCLUDED
#include "svg_basics.h"
#include "svg_parser.h"
#include "svg_styleset.h"
namespace svg
{
//////////////////////////////////////////////////////////////////////
// reference: forward delclaration
//////////////////////////////////////////////////////////////////////
#ifndef SVG_ELEMENT_INCLUDED
class	element;
#endif

//////////////////////////////////////////////////////////////////////
// SVG request: a message, that "bubbles" across the SVG graph.
// Using specialized children classes of this structure, the root
// element makes its children to do something useful: 
//	- to initialize after load (an initialize request)
//	- to measure its size (a size request)
//	- to render on a picture (a canvas request), 
// 
//	In analogy to an window system, each SVG graph element has 
//	an virtual method to handle the received request: on_request.
//	It is as important method, as the "message switch" in the core 
//	Windows application! In fact it determines the bhaviour of the
//	element.
//////////////////////////////////////////////////////////////////////
struct request  
{
	public:
    enum kind
    {
			none				= 0, //just to mark an empty value
			initialize			= 1, //just to mark the first index of the style
            measure				= 2, //measure the drawing
			override			= 3, //override an attribute

			//... other preparation requests:
            render				= 100,
			filter				= 101,

			//...
			//to_root requests:
			to_root_base		= 1000, 
			prepare				= 1000, //just assign the elements and attribute storage 
			get_style			= 1001,
			prepare_measure		= 1002,
			prepare_render		= 1003
			//...
    };

    public:
		//Basic constructor
		//Arguments:
		//	type:	request type (accordig the enumertion)
		//	attrs:	attribute storage, that contains the element's attributes
		//	elems:	element storage, that contains the SVG graph elements
		request(kind type, attr_storage &attrs, element_storage &elems): 
							m_type(type), m_attrs(&attrs), m_elems(&elems) {}
        virtual ~request() {}

		//Simplified constructor - for the "to_root" requests
		//Arguments:
		//	type:	request type (accordig the enumertion)
		request(kind type): m_type(type), m_attrs(NULL), m_elems(NULL) {}
	 
		//And supplemental method that enables to place the 
		//information about the graph storage later on:
		//Arguments:
		//	attrs:	attribute storage, that contains the element's attributes
		//	elems:	element storage, that contains the SVG graph elements
		void assign(attr_storage &attrs, element_storage &elems)
		{
			m_attrs = &attrs; m_elems = &elems;
		}
		//////////////////////////////////////////////////////////////////////
		// Information, what is it, and its capabilities
		//////////////////////////////////////////////////////////////////////
		//Returns kind of this request.
		kind type() {return m_type;}

		//Returns true, when request can be queried for elements or attributes
		//(Some requests cannot)
		bool can_query() {return (m_attrs && m_elems);}

		//Returns false, when the request cannot be forwarded to element's
		//descendants (all "to_root" requests have it on)
		bool can_forward() {return (m_type >= to_root_base) ? false : true;}

		//////////////////////////////////////////////////////////////////////
		// Elements and  attributes handling - utilites for the SVG graph
		//////////////////////////////////////////////////////////////////////
		//Returns the attribute, associated with a given handle
		//Arguments:
		//	id:		a handle to an attribute 
		attribute *get_attribute(handle id) 
			{ if (m_attrs && id < (*m_attrs).size()) return &((*m_attrs)[id]); else return NULL;}

		//Returns an element, assigned to a handle
		//Arguments:
		//	id:				a handle to an element, obtained from get_id()
		//	expected_class:	optional: the name of the expected class ("text", "ext:earray", etc.)
		//					used for rudimentary type checking
		element  *get_element(handle id, const char *expected_class = NULL);
        
    protected:
		enum kind					m_type;
		//these are pointers below, because they can be initialized later 
		//(or -for some kinds of requests - even never)
		attr_storage				*m_attrs;
		element_storage				*m_elems;
}; //request

//////////////////////////////////////////////////////////////////////
// req_get_style: one of inquiry requests, sent from an element to 
// the root of SVG graph. Retuns the actual style.
// Used by an element during font attributes resolution, on creation 
// of SVG graph.
//////////////////////////////////////////////////////////////////////

struct req_get_style: public request
{
	public:
		//Constructor
		//The svg::root element should assign the SVG graph attrrs and elems
		//storages! Then it will come back to the sender.
		req_get_style(): request(request::get_style) {}
	
	public:	//public fields - can be modified 
		styleset	style;	//Each elements modify this style according 
							//its own settings
}; //req_get_style

//////////////////////////////////////////////////////////////////////
// req_prepare_request: one of requests, sent from an element to 
// the root of SVG graph, - a "request to prepare working reqest"
// for the measurement or rendering
// Used by elements for gradients, patterns or blur effects
// when they have to render something "on their own"
//////////////////////////////////////////////////////////////////////

struct req_prepare: public request
{
	public:
		//Constructor
		//Arguments:
		//	type:		request type (accordig the enumertion), for example:
		//				request::prepare_measure
		//				request::prepare_render
		//
		//	content:	request, to be updated by the root (should be of proper type)
		//The svg::root element should assign the SVG graph attrrs and elems
		//storages! Then it will come back to the sender.
		req_prepare(kind type, request &content): request(type), req(content) 
		{}
	
	public:	//public fields - can be modified 
		request	&req;	//Each elements modify this style according 
							//its own settings
}; //req_prepare_request

#ifdef TEX_EXTENSIONS
//////////////////////////////////////////////////////////////////////
// req_override: a request, which overrides an attribute in
// the element, which receives it.
// It is always send directyly to the target, not to all
// Used by tex:: extensions, to manage a couple of SVG drawings
//////////////////////////////////////////////////////////////////////
struct req_override: public request
{
	public:
		//Constructor
		//Arguments:
		//	propagate_to_children:		true, when the children should be updates
		req_override(bool propagate_to_children): request(override), 
		propagate(propagate_to_children)
		{}
	
	public:	//public fields - can be modified 
		attribute	value;		//A new attribute, which should override its previous 
								//version in all places, where it exists.
								//Overriden when the request is already created
		bool		propagate;	//True, when it should override it in all the children
								//of the target element
}; //req_prepare_request
#endif TEX_EXTENSIONS
}; //namespace svg
#endif SVG_REQUEST_INCLUDED
