//----------------------------------------------------------------------------
// SVG Viewer - Version 1.0
// A library of SVG renderer
// Copyright (C) 2008 Witold Jaworski
// Contact:	witold.jaworski@airplanes3d.net
//			http://airplanes3d.net
// 
// Contains elements from:
//			AGG - a graphic library (Copyright Maxim Shemanarev, mcseem@antigrain.com)
//          http://antigrain.com
// Uses functions from:
//          Expat - a XML parser library (Copyright James Clark)
// 
// SVG Viewer is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
// 
// SVG Viewer is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with SVG Viewer; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, 
// MA 02110-1301, USA.
//----------------------------------------------------------------------------

#ifndef SVG_ROOT_INCLUDED
#include "svg_group.h"
#include "svg_req_render.h"
#define SVG_ROOT_INCLUDED

namespace svg
{
//////////////////////////////////////////////////////////////////////
// root: the first element of the SVG graph
//////////////////////////////////////////////////////////////////////
class root : public group 
{
	public:
		//////////////////////////////////////////////////////////////////////
		// constructor and overriden virtual methods
		//////////////////////////////////////////////////////////////////////

		root(): group(NULL, 0),m_bounds(0,0,-1,-1), m_nominal(0,0,-1,-1), m_vertices(0)
		{}

		//Initializes the SVG graph, using content of a SVG file
		//Arguments:
		//	fileaname:	name of an SVG file
		void load(const char *filename)
		{
			parser prs(m_paths, m_attrs, m_elems);
			on_attribute(prs,"id","~root"); //register itself
			prs.parse(filename);
			commit(prs);
			//m_paths.arrange_orientations_all_paths(agg::path_flags_ccw);
			//send the initial requests:
			initialize();
			measure();
		}

		virtual const char * tag() {return "root";};

		virtual ~root()
		{	//Do it to avoid recursion deadlock:
			if (is_valid(m_id)) m_elems[m_id].assign(NULL); 
		}

		//////////////////////////////////////////////////////////////////////
		// handling requests from the descendants :
		//////////////////////////////////////////////////////////////////////
		//request handler: receives the "to_root" requests
		virtual handle on_request(request &req) 
		{
			switch(req.type())
			{
			case request::prepare: //general prepare:
				if(!req.can_query()) ((req_prepare &) req).req.assign(m_attrs, m_elems);
				return skip(req);

			case request::prepare_render:
			case request::prepare_measure:
				if(!req.can_query()) prepare_geometry((req_prepare &) req);
				return skip(req);
				
			case request::get_style:
			default: 
				//assign the SVG attribute storage to it - then they
				//will process of this request (see element::to_root() code)
				if (!req.can_query()) req.assign(m_attrs,m_elems);
				//(can_query() == false when the m_attrs and m_elems are not assigned)
				//We are chcecking this flag to avoid replacement of the storage
				//information by some "super elements", located on as the parents
				//of this element (in case of use an xml files, that defines a
				//compound of some SVG pictures)
				return skip(req);
			}//switch
		}

		//////////////////////////////////////////////////////////////////////
		// Other methods and fields
		//////////////////////////////////////////////////////////////////////
		//number of the vertices
		unsigned vertices() {return m_vertices;}
		
		//measured size of all elements
		rectangle bounds() {return m_bounds;}
		
		//declared size of the picture
		rectangle extents() {return m_nominal;}

		//width and the height of the nominal picture
		double width() {return m_nominal.x2 - m_nominal.x1;}
		double height() {return m_nominal.y2 - m_nominal.y1;}

		//Renders the final image
		//Arguments:
		//	buffer:		the buffer for the rendered image;
		//	clip:		clipping bounds (after transformation)
		//	trans:		initial transformation (applied to the whole image)
		//	bkgnd:		background color
		//	opacity:	optional: overall opacity of the rendered picture.
		//	expand:		offset of the filled shapes: 0.0, when the shape has not 
		//				to be thinned/fatted
		//	gamma:		optional: gamma correction, used in antialiasing
		void render(pixfmt buffer, rectangle clip, transform trans,	rgba8 bkgnd, 
					double opacity = 1.0, double expand = 0.0, double gamma = 1.5)
		{
			req_render req(buffer, clip, bkgnd, m_paths,m_attrs,m_elems, trans);
			req.gamma(gamma);
			req.expand(expand);
			forward(req);
		}

		//Renders the final image - simpler version
		//Arguments:
		//	buffer:		the buffer for the rendered image;
		//	trans:		initial transformation (applied to the whole image)
		//	bkgnd:		background color
		//	opacity:	optional: overall opacity of the rendered picture.
		//Renders the picture in the nominal size
		void render(pixfmt buffer, transform trans, rgba8 bkgnd, double opacity = 1.0)
		{
			render(buffer, m_nominal, trans, bkgnd, opacity); //no special gamma, nominal size
		}

	protected:

		//Initializes the SVG graph (called right after the creation)
		void initialize()
		{
			request req = request(request::initialize,m_attrs,m_elems);
			forward(req);
		}

		//Some statistics about the SVG graph (called just after initialize()) 
		void measure()
		{
			transform empty; //neutral value, just to fill the 4th argument
			req_measure req(m_paths,m_attrs,m_elems,empty);
			forward(req);
		
			m_bounds = req.bounds;
			if (req.width == 0 || req.height == 0)
			{ //no declaration of the size? nominal = measured
				m_nominal.x1 = m_bounds.x1;
				m_nominal.y1 = m_bounds.y2;
				m_nominal.x2 = m_bounds.x2;
				m_nominal.y2 = m_bounds.y2;
			}
			else //declared nominal size always has a 0,0 orgin: 
			{
				m_nominal.x1 = 0;
				m_nominal.y1 = 0;
				m_nominal.x2 = req.width;
				m_nominal.y2 = req.height;
			}

			m_vertices = req.vertex_count;
		}
		//prepares a geometry (measure, render) request:
		//handles the answers to render::prepare_measure,
		//						 render::prepare_render
		void prepare_geometry(req_prepare &req)
		{
			((req_geometry &) req.req).assign(m_paths,m_attrs,m_elems,transform());
		}

	protected:
	//storage of the SVG graph elements
		path_storage				m_paths;
		attr_storage				m_attrs;
		element_storage				m_elems;

	//SVG graph summary (for the potential "client" - returned by the functions
	//like bounds 
		rectangle					m_nominal;	//nominal (declared) size	
		rectangle					m_bounds;	//measured size
		unsigned					m_vertices;
		
}; //root

}; //namespace svg
#endif SVG_ROOT_INCLUDED
