//----------------------------------------------------------------------------
// SVG Viewer - Version 1.0
// A library of SVG renderer
// Copyright (C) 2008 Witold Jaworski
// Contact:	witold.jaworski@airplanes3d.net
//			http://airplanes3d.net
// 
// Contains elements from:
//			AGG - a graphic library (Copyright Maxim Shemanarev, mcseem@antigrain.com)
//          http://antigrain.com
// Uses functions from:
//          Expat - a XML parser library (Copyright James Clark)
// 
// SVG Viewer is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
// 
// SVG Viewer is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with SVG Viewer; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, 
// MA 02110-1301, USA.
//----------------------------------------------------------------------------

#ifndef SVG_SEGMENT_INCLUDED
#include "svg_basics.h"
#include <math.h>
#define SVG_SEGMENT_INCLUDED

namespace svg
{
//////////////////////////////////////////////////////////////////////
// segment: it is a helper object: a fragment of the straight, 2D line
//		    (Used in the positioning text on a curve, and other places)
// It contains the parameters of the 2D line: 
//	- coordinates of its "origin": x0,y0 - a single point, where t = 0
//	- direction: dx,dy - a unit vector of the line direction
//	- t: length of the line
// Segment can give you the end point coords - using the point() function.
// In degenerated from length of the segment can be zero (t=0). In this
// case it does not matter, which direction dx,dy it has - always evaluates
// to the origin(). In case of svg::conv_locator, it is very useful form
// to represent the first point of the first curve segment.
//////////////////////////////////////////////////////////////////////
class segment  
{
	public:
		//////////////////////////////////////////////////////////////////////
		// constructor 
		//////////////////////////////////////////////////////////////////////
		//Creates a new segment from two endpoints
		//	Arguments:
		//	x1, y1: first point of the segment
		//	x1, y2: last point of the segment
		segment(double x1, double y1, double x2, double y2)
		{
			m_x0 = x1; m_y0 = y1;
			double dx = x2 - x1, dy = y2 - y1;
			m_t = sqrt(dx*dx + dy*dy);

			if (is_valid()) //checks the value of the m_t: is it zero? 
			{
				m_dx = dx/m_t; m_dy = dy/m_t; //m_dx, m_dy: an unit vector
			}
			else
			{
				m_t = 0;
				m_dx = 1.0; m_dy = 0; //the direction does not matter when 
									  //the length is zero...
			}
		}

		//Other way of passing the segment parameters
		//	Arguments:
		//	x, y: first point of the segment
		//	dx, dy: direction of the line (can be any vector - it will be normalized)
		//	t		:advance coefficent. Can be negative. For dx,dy as unit vector
		//			 it is also the length of the segment.
		segment(double x, double y, double t, double dx, double dy)
			{m_x0 = x; m_y0 = y; m_t = t; m_dx = dx; m_dy = dy;}

		//Creates a degenerated segemnt at given point
		//	Arguments:
		//	x, y: the origin of the segment
		segment(double x, double y)	{ *this =	segment(x, y, x, y) ; }

		//Creates absolutly degenrated segment
		segment() { *this = segment(0,0);}		
		
		//Other way of passing the segment parameters
		//	Arguments:
		//	start:	the end point of this segment is the start point 
		//	end:	the end point of this segment is the end point 
		segment(segment &start, segment &end)
			{*this = segment(start.x(), start.y(), end.x(), end.y()); }
	public:
		//////////////////////////////////////////////////////////////////////
		// Other methods
		//////////////////////////////////////////////////////////////////////
		//Access to the field values:
		inline double x0() {return m_x0;}
		inline double y0() {return m_y0;}
		inline double dx() {return m_dx;}
		inline double dy() {return m_dy;}
		inline double length() {return m_t;}

		//Returns true, when the segment is not degenerated
		inline bool is_valid() {return fabs(m_t) > SVG_EPSILON;}
		
		//coordinates of any point on the line 
		//Arguments: 
		//	t:	advance of the desired fragment 
		//		(for 0.0 <= t <= m_t: within segment)
		//		can be nagative!
		double x(double t) {return m_x0 + t * m_dx;}
		double y(double t) {return m_y0 + t * m_dy;}
		
		//segment's end point coordinates
		double x()	{return x(m_t);}
		double y()	{return y(m_t);}
		//
		void point(double *px, double *py, double at = 0.0) 
													{*px = x(at); *py = y(at);}
		//Calculates the angle (in radians) between this segment and another
		//Arguments:
		//	s:	another segment
		//Returns a value from -pi to pi
		double angle(segment &s) {	return angle(s.dx(), s.dy());	}

		//Calculates the angle (in radians) between this segment and global X axis
		//Returns a value from -pi to pi
		double angle()	{return angle(1.0, 0.0);}

	protected:

		//Calculates the angle (in radians) between this segment and another
		//Arguments:
		//	nx:	normal vextor - x coordinate
		//	ny: normal vector - y coordinate
		//Returns a value from -pi to pi
		double angle(double nx, double ny)
		{
			double sinx = nx * m_dy  -  ny * m_dx; 
			double cosx = nx * m_dx  +  ny * m_dy; 
			if(sinx > 0.0) //Upper half of the arc
			{
				if (cosx > 0.0) //right quarter
				{
					return asin(sinx);	
				}
				else			//left quarter
				{
					return agg::pi - asin(sinx);
				}
			}
			else //Lower half of the arc
			{
				if (cosx > 0.0) //right quarter
				{
					return asin(sinx);	//returns an negative value
				}
				else			//left quarter
				{
					return - agg::pi - asin(sinx); //asin(sinx) is always > -pi, here
				}
			}
		}

	protected:
		//////////////////////////////////////////////////////////////////////
		// Instance fields
		//////////////////////////////////////////////////////////////////////
	double m_x0, m_y0; //origin
	double m_dx, m_dy; //direction (an unit vector)
	double m_t;		   //length of the segment	
}; //segment

}; //namespace svg
#endif SVG_SEGMENT_INCLUDED
