//----------------------------------------------------------------------------
// SVG Viewer - Version 1.0
// A library of SVG renderer
// Copyright (C) 2008 Witold Jaworski
// Contact:	witold.jaworski@airplanes3d.net
//			http://airplanes3d.net
// 
// Contains elements from:
//			AGG - a graphic library (Copyright Maxim Shemanarev, mcseem@antigrain.com)
//          http://antigrain.com
// Uses functions from:
//          Expat - a XML parser library (Copyright James Clark)
// 
// SVG Viewer is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
// 
// SVG Viewer is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with SVG Viewer; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, 
// MA 02110-1301, USA.
//----------------------------------------------------------------------------
#include "svg_exception.h"
#include "svg_shape.h"
#include "svg_use.h"
#include "svg_req_render.h"
#include "svg_filter.h"
#include "svg_opacity.h"
#include "svg_painter.h"

namespace svg
{
//----------------------------------------------------------------------------
	void shape::on_attribute(parser &prs, const char * name, const char *value)
	{
		if (strcmp(name, "transform") == 0) 
		{	
			m_trans = prs.parse_transform(value);
		}
		else
		if (strcmp(name, "display") == 0) 
		{	
			m_display = !(strcmp(value,"none") == 0);
		}
		else
		if (strcmp(name, "filter") == 0) 
		{	
			m_filter = prs.get_id(value); //id of the filter element
		}
		else
		if (strcmp(name, "mask") == 0) 
		{	
			m_mask = prs.get_id(value); //id of the filter element
		}
		else
		if (strcmp(name, "clip-path") == 0) 
		{	
			m_clip_path = prs.get_id(value); //id of the filter element
		}
		if (strcmp(name, "opacity") == 0) 
		{	//Opacity value is stored (as a filter) in an unnamed element
			m_opacity = opacity::create(prs,this,level()+1,prs.to_double(value));
		}
		else //try to store it as one of the style attributes:
			if(! prs.store_as_style(*this, name, value)) 
			{//if the style is not recognized:
				element::on_attribute(prs,name,value);
			}
	}

	//Returns actual bounding box (required for gradients or paterns
	//with units set to objectBoundingBox).
	//Sets the internal m_bounds field, on first use.
	rectangle shape::bounding_box()
	{
		if(!m_bounds.is_valid())
		{
			transform identity = transform();
			m_bounds = bounding_box(identity, true);
		}
		return m_bounds;
	}//shape::bounding_box

	//Returns extents of a transformed shape
	//Arguments:
	//	trans:			transformation, which should be applied before
	//					the measurement
	//	force_measure:	when true, forces to measure (used in bounding_box(),
	//					to avoid endless recursion)
	rectangle shape::bounding_box(transform trans, bool force_measure)
	{
		rectangle e;

		if(!force_measure && svg::is_translation_only(trans))
		{//small optimisation: do not measure itself, when you can get
		 //the required information in a cheaper way:

			e = bounding_box(); //just take the nominal coordinates...
			//.. and move them by trans:
			trans.transform( &(e.x1), &(e.y1));
			trans.transform( &(e.x2), &(e.y2));
		}
		else //no easy way: measure itself!
		{
			req_measure measure;
			//make the measure request usable:
			request req = req_prepare(request::prepare_measure, measure);
			to_root(req);
			if(measure.can_query())
			{
				transform buf = m_trans; //preserve actual transformation
				m_trans = trans;	 //set the transform
				on_request(measure);
				e = measure.bounds;
				m_trans = buf; //restore current transformation
			}//end if
		}
		return e;
	}//shape::bounding_box

	//Some details of the style or referenced objects have to be prepared
	//before each rendering. This procedure does it.
	//Arguments
	//	req:	render request (just to retrieve elements or attributes)
	//	style:	the resulting styleset, that will be used in this rendering 
	void shape::prepare_render(request &req, styleset &style)
	{
		
		if(req.type() != request::render) return; //We serve render requests, only!

		if( style.of(attribute::fill_mode).val.draw_mode == other
			&&
			style.has(attribute::fill_painter) 
		  )
		{
			painter *paint = painter::from(style,req,attribute::fill_painter);
			if (paint && paint->box_required()) 
			{
				rectangle box = bounding_box();
				paint->set(box);
			}
			else
			{
				paint->reset();
			}
		}

		if( style.of(attribute::stroke_mode).val.draw_mode == other
			&&
			style.has(attribute::stroke_painter) 
		  )
		{
			painter *paint = painter::from(style,req,attribute::stroke_painter);
			if (paint && paint->box_required()) 
			{
				rectangle box = bounding_box();
				paint->set(box);
			}
			else
			{
				paint->reset();
			}
		}
	}//end prepare_render


	//request handler:
	handle shape::on_request(request &req)
	{
		switch(req.type())
		{

#ifdef TEX_EXTENSIONS
		case request::override:
			{
				req_override & rovr = (req_override &) req;
				switch(rovr.value.type)
				{
				case attribute::filter_ref:
						m_filter = rovr.value.val.id;
						break;

				case attribute::mask_ref:
						m_mask = rovr.value.val.id;
						break;

				case attribute::clip_path_ref:
						m_clip_path = rovr.value.val.id;
						break;

				case attribute::opacity_ref:
						m_opacity = rovr.value.val.id;
						break;

				case attribute::display:
						m_display = rovr.value.val.boolean;
						break;
				
				default:
					element::on_request(req);
				}
				return skip(req);
			}
#endif TEX_EXTENSIONS

		case request::get_style:
			{
				req_get_style & sreq =  (req_get_style &) req;
				return get_style(sreq);
			}	
		
		case request::measure:
			{
				return render(req);
			}

		case request::render:
			{
				req_render &rnd = (req_render &)req;

				/*	
					There are some composite operations, triggered by
					some common XML SVG attributes of the shape object:
					 - opacity
					 - clip path
					 - mask
					 - filter

					Each of them renders the result to an intermediate
					picture buffer, which is used by the next operation.
					When a shape (it can be any drawing object, also a group)
					has two or more such attributes, they are applied in following 
					order:
					1. The basic shape, without any composite operations, is rendered
					   to an intermediate buffer;
					2. A filter is applied to result of 1. and stored in a new buffer
					3. A mask is applied to result of 2. and stored in a new buffer
					4. A clip path is applied to result of 3. and stored in a new buffer
					5. Result of 4. is applied to the final result picture with 
					   specified opacity

					For a basic shape, with no composite attributes, operation 1
					is performed, storing the result to final result picture.
					The conditional clauses below implements the hierarchy described
					above.

					Each composite operation is implemented by apropriate svg::filter
					class, called in a helper procedure, for code readibility.
					Each operation renders
					The "switching off" the subsequent filters along steps 1..5 is
					controlled by req_render::restricted_filter property. 
					This object will receive one render request for each of 1.. 5 
					operation. The req_render::restricted_filter property allows 
					shape to determine, which features should not be applied on
					the current stage (for example: if actually restricted filter
					is the mask, no clip path, nor opacity should be applied). 
					   
				*/

				if (is_valid(m_filter) && m_filter == rnd.restricted_filter)
				{
					return render(rnd); //no filter, no mask, no clip 
				}
				else
				{	
					if (is_valid(m_mask) && m_mask == rnd.restricted_filter)
					{
						return render_with_filter(rnd); //no mask, no clip 
					}
					else
					{	
						if (is_valid(m_clip_path) && m_clip_path == rnd.restricted_filter)
						{
							return render_with_mask(rnd); //no clip 
						}
						else
						{	
							if (is_valid(m_opacity) && m_opacity == rnd.restricted_filter)
							{
								return render_with_clip_path(rnd); //no opacity 
							}
							else
							{	
								return render_with_opacity(rnd); //render with all options
							}				
						}				

					}				
				}
								
			}

		case request::prepare_render:
		case request::prepare_measure:
			{
				req_geometry & greq = (req_geometry &) ((req_prepare &) req).req;
				return get_style(greq);
			}
		default: 
			return element::on_request(req);
		}//switch
	}

	//Helper function, used to obtain an eventual filter object (i.e a clip path,
	//a mask, a filter). The svg::source_graphic_buffer (see svg_req_filter.cpp)
	//is responsible for setting proper restriction flags in the request structure.
	//Arguments:
	//	id:			 handle to eventual filter
	//	req:		 request (req_render) structure
	static filter *get_filter(handle id, request &req)
	{
		req_render &rnd = (req_render &) req;

		if (!is_valid(id) || id == rnd.restricted_filter)
		{
			return NULL; //do not use this filter
		}
		else
		{
			filter *fltr = NULL;
			if (is_valid(id))  //try to get the filter object:
			{
				fltr = (svg::filter *) req.get_element(id, "filter");
			}
	
			if (fltr && fltr->initialized())
			{
				return fltr;
			}
			else
			{
				return NULL;
			}
		}
	}// get_filter

	//Helper procedure to render just with a filter, or normal
	handle shape::render_with_filter(request &req)
	{
		svg::filter *fltr = get_filter(m_filter,req);
		if (fltr) 
		{	
			fltr->perform(*this,(req_render &) req); //filter this!
			return skip(req);
		}
		else
		{
			return render(req);
		}
	}

	//Helper procedure to render just with a mask or a filter
	handle shape::render_with_mask(request &req)
	{
		svg::filter *msk = get_filter(m_mask,req);
		if (msk) 
		{	
			msk->perform(*this,(req_render &) req); //filter this!
			return skip(req);
		}
		else
		{
			return render_with_filter(req);
		}
	}

	//Helper procedure to render just with a clip path, a mask, or a filter
	handle shape::render_with_clip_path(request &req)
	{
		svg::filter *clp = get_filter(m_clip_path,req);
		if (clp) 
		{	
			clp->perform(*this, (req_render &) req); //filter this!
			return skip(req);
		}
		else
		{
			return render_with_mask(req);
		}
	}

	//Helper procedure to render just with a opacity, clip path, a mask, or a filter
	handle shape::render_with_opacity(request &req)
	{
		svg::opacity *op = (svg::opacity *) get_filter(m_opacity,req); //the explicite type
		if (op)												 //is important here,					
		{	
			op->perform(*this, (req_render &) req); //because the opacity::perform()
													//overrides the default
													//filter::perform method!
			return skip(req);
		}
		else
		{
			return render_with_clip_path(req);
		}
	}

} // namespace svg
