//----------------------------------------------------------------------------
// SVG Viewer - Version 1.0
// A library of SVG renderer
// Copyright (C) 2008 Witold Jaworski
// Contact:	witold.jaworski@airplanes3d.net
//			http://airplanes3d.net
// 
// Contains elements from:
//			AGG - a graphic library (Copyright Maxim Shemanarev, mcseem@antigrain.com)
//          http://antigrain.com
// Uses functions from:
//          Expat - a XML parser library (Copyright James Clark)
// 
// SVG Viewer is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
// 
// SVG Viewer is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with SVG Viewer; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, 
// MA 02110-1301, USA.
//----------------------------------------------------------------------------

#ifndef SVG_TEXT_INCLUDED
#include "svg_shape.h"
#include "svg_tspan.h"
#include "svg_group.h"
#include "svg_linear_text_position.h"
#define SVG_TEXT_INCLUDED

namespace svg
{
//////////////////////////////////////////////////////////////////////
// text: a text content in the SVG graph
//////////////////////////////////////////////////////////////////////
class text : public group
{
	public:
		//////////////////////////////////////////////////////////////////////
		// constructor and overriden virtual methods
		//////////////////////////////////////////////////////////////////////

		text(element *parent, unsigned level): group(parent, level) 
		{
			m_inner_tspan = NULL;
		}

		virtual const char * tag() {return "text";};

		virtual void on_attribute(parser &prs, const char * name, const char *value)
		{
			if (
					(strcmp(name, "x") == 0) 
				||
					(strcmp(name, "y") == 0) 
				||
					(strcmp(name, "dx") == 0) 
				||
					(strcmp(name, "dy") == 0) 
				||
					(strcmp(name, "rotate") == 0) 
			   )
			{	
				if (!m_inner_tspan) 
				{
					add_tspan(prs); //initializes m_inner_tspan
				}
				if (m_inner_tspan) 
				{
					m_inner_tspan->on_attribute(prs, name,value);
					m_inner_tspan->commit(prs); //usually each attribute creates 
												//a new svg::earray element!
				}
			}
			else 
			{//if the style is not recognized:
				group::on_attribute(prs,name,value);
			}
		}//on_attribute

		virtual void on_content(parser &prs, const char* s, int len) 
		{
			if (commited() == 0) commit(prs); //just to catch the first m_id_last
			//This element will receive another commit() call, at the end of 
			//its xml definition (from svg::parser.end_element())

			if (!m_inner_tspan) 
			{
				add_tspan(prs,s,(unsigned) len);
				//In this case it has to be already commited: 
				if (m_inner_tspan) m_inner_tspan->commit(prs);
			}
			else
			{
				//m_id_last: number of prs.elements when first inner tspan was created
				if(prs.elements() == (m_id_last + 1)) //there are no more internal objects, yet
				{
					m_inner_tspan->append(s,(unsigned) len);
				}
				else //there are other tspans: add another!
				{
					add_tspan(prs,s,(unsigned) len);
				}
			}
		}

		//Default implementation of the rendering: linear text...
		virtual handle render(request &rreq)
		{
			if (!display()) return skip(rreq);

			req_geometry &req = (req_geometry &) rreq;

			linear_text_position tpos;
			return render(req, tpos);
		}

	public:
		//////////////////////////////////////////////////////////////////////
		// Other methods
		//////////////////////////////////////////////////////////////////////
		//Method used by svg::tref element, to get the content of a referenced
		//text.
		tspan *content() {return m_inner_tspan;}

	protected:
		//version for child classes (text_path, for example), when the current
		//text poition is already set to a curve, for example)
		handle render(req_geometry &req, text_position &tpos)
		{
			if (!req.can_forward()) return skip(req);
		
			//preserve received transformation and style
			transform	trans	= req.trans;	
			transform	ptrans;  //used to deliver the content to the temporary 
								 //req.painter_trans
			styleset	style	= req.style;
			
			//change the request, according styles and transformation of this group
			req.trans.premultiply(m_trans);		
			ptrans = req.trans; //for the painters, use this value!
			req.painter_trans = &ptrans;

			req.style.update(req,m_attribute_first, m_attribute_last);

			//some final preparation: 
			//bounding box is passed to eventual painters, 
			shape::prepare_render(req, style);


			//it may happen, that m_inner_tspan has smaller id than this->id:
			if (m_inner_tspan && m_inner_tspan->id() < m_id)
			{//in such case we have to call it separately:
			 //to let it set the current text position - tpos.  
				m_inner_tspan->render(req,tpos);
			}

			//different way of forwarding: <tspan>s elements are treated
			//in the separate way
			handle id = first_child();
			
			while(id <= m_id_last)
			{
				element *elm = req.get_element(id);
				//do not render the break or forward used elements
				if (!elm) break;
				//when it should not be used at this moment: skip
				if (elm && !is_valid(elm->forward_use())) 
				{ //OK, let's render:
					if (
							strcmp(elm->tag(),"tspan") == 0
							||
							strcmp(elm->tag(),"tref") == 0
					   )
					{
						id = ((tspan *)elm)->render(req, tpos);		
					}
					else
					{
						id = elm->on_request(req); 
					}//if
				}
				else 
				{
					id++; //next id, please....
				}
			}//while

			//restore previous state of the request - (bounds) transformation and style
			req.painter_trans = NULL;
			req.trans	= trans; 
			req.style	= style;

			//return the result, received from the descendants
			return id;
		}

		//creates and stores an unnamed tspan, which will contain
		//the textual data from internal content of the xml <text>
		//element. If it is a first such call - it initializes the
		//m_inner_tspan field.
		void add_tspan(parser &prs, const char *utf8 = NULL, unsigned len = 0)
		{
			element *tp = tspan::create(prs, this, level()+1, 
											utf8, len, (m_inner_tspan != NULL));
			if(!m_inner_tspan) m_inner_tspan = (tspan *) tp;
		}
		
	protected:
		//////////////////////////////////////////////////////////////////////
		// Instance additional fields
		//////////////////////////////////////////////////////////////////////
		//m_inner_tspan is just a last svg::tspan, that has been created to handle the 
		//text inner content, It can be the last from many similar tspans, created
		//to match the content of the <text> element. It can also conatin no text 
		//at all - being just a placeholedr for the initial current_text_pos(ition)
		//settings
		tspan *m_inner_tspan;

}; //text

}; //namespace svg
#endif SVG_TEXT_INCLUDED
