//----------------------------------------------------------------------------
// SVG Viewer - Version 1.0
// A library of SVG renderer
// Copyright (C) 2008 Witold Jaworski
// Contact:	witold.jaworski@airplanes3d.net
//			http://airplanes3d.net
// 
// Contains elements from:
//			AGG - a graphic library (Copyright Maxim Shemanarev, mcseem@antigrain.com)
//          http://antigrain.com
// Uses functions from:
//          Expat - a XML parser library (Copyright James Clark)
// 
// SVG Viewer is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
// 
// SVG Viewer is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with SVG Viewer; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, 
// MA 02110-1301, USA.
//----------------------------------------------------------------------------

#ifndef SVG_TEXT_POSITION_INCLUDED
#include "svg_font.h"
#define SVG_TEXT_POSITION_INCLUDED

namespace svg
{
	typedef agg::pod_bvector<unicode> string; //used by all text elements
											  //to store any text	
//////////////////////////////////////////////////////////////////////
// text_position: it is a helper object: something, which in the SVG
//				  specification is called "current text position"
// It is a base class for two descendants: 
//		- linear_text_position (for standard texts)
//		- curved_text_position (for the "a text on the path" effect) 
//
// The text elements use one of these two child classes to render the text. 
//////////////////////////////////////////////////////////////////////
class text_position  
{
public:
	//constructor: creates a text position in an uninitioalized state
	text_position()
	{
		m_text = NULL;
		m_base = m_pos = m_count = 0;
		m_glyph = NULL;
		m_scale = 1.0;
		m_anchor = 0.0;
		m_wmode = lr;
		m_valid = false;	//true, if the text position has been properly 
		m_auto_kerning = false;
		m_kerning_value = 0.0;
	}
	//Applies some style attributes to the current text position
	//Arguments:
	//	req:	a request - to obtain the attribute data;
	//	style:	an effective (accumulated from the parents) style,
	//			which should contain the font definition
	//This method tries to find and recognize following style attributes:
	//	attribute::font:
	//	attribute::font-size:
	//	attribute::writing_mode
	//	attribute::text_anchor
	//and uses them to fill the internal data fields of this class
	//It also modifies the style: scales the stroke-width attribute 
	//to match the internal transform of the glyph vertex source!
	void apply(request &req, styleset &style);

	//Sets the initial current text position
	//Arguments:
	//	x,y:		coordinates of the starting position (for the text 
	//				on a curve the "y" coordinate is ignored)
	//	text:		the text, which fragment will be drawn;
	//	pos:		position of the first character that belonggs to this
	//				text fragment (text chunk)
	//	count:		number of characters to write (length of the text chunk)
	//	dx,dy:		offset from current text position, that has to be applied
	//				to each character within this text chunk
	//	rotation:	additional rotation, applied individually to each character
	//				within this text chunk
	virtual void set(double x, double y, string &text, unsigned pos, 
					unsigned count, double dx, double dy, double rotation);
	
	//Current text position on the drawing:
	virtual double x() = 0; 
	virtual double y() = 0; 

	//actual glyph transformation
	virtual transform transformation() = 0;

	//advances text position to the next point (and to next character 
	//within text chunk) returns false, when passed the end of the text 
	//chunk declared in the set() method 
	virtual bool advance() = 0; 

	//return true, when text position has been set up and is ready to use
	//(the child classes are responsible for the state of m_valid field)
	bool is_valid() {return m_valid;}
	
	//returns the geometry (path) for the actual glyph
	vsource *outline() {return m_glyph ? m_glyph->vertices() : NULL;}
	
	//first character in this text chunk
	unsigned first() {return m_base;}

	//number of characters within this text chunk
	unsigned count() {return m_count;}


protected:

	
	//offset X to the new character position:
	double advance_x();

	//offset Y to the new character position:
	double advance_y();

	//actual kerning in the Y direction (to the previous character):
	double kerning_x() 
		{return (m_glyph && m_auto_kerning) ? m_glyph->kerning_x() : m_kerning_value;};

	//actual kerning in the Y direction (to the previous character):
	double kerning_y() 
		{return (m_glyph && m_auto_kerning) ? m_glyph->kerning_y() : m_kerning_value;};

	//Internal glyph transformation: offset from the reference point: 
	//	- for "left-right": baseline (y=0), on left edge of the glyph box;
	//	- for "top-bottom": middle of the top edge of the glyph box;
	// into natural center of the glyph (identified as the point 
	// 0,0 within the glyph bounds).
	//NOTICE: No scaling! - it is all in the natural font coordinates!
	transform internal_transform();

	//returns the ovearll span of actual text chunk: 
	//Returns a rectangle, where: 
	//	x2 => total offset of the text position along x direction
	//	y2 => total offset of the text position along y direction
	//Beware, that offset of the text position may be slightly different
	//than the overall text size - it does not take into account the
	//real glyph boxes, only the advances between subsequent characters
	rectangle offset();

	//A character code (an unicode) at given text position
	//Arguments:
	//	index: index of the code, within actual text chunk
	//		   (0 <= index < count) 
	unicode code(unsigned index)
	{	//we simulate "left to right" direction by reading the text
		//backwards:
		if (m_wmode == rl) index = m_count - (index + 1);
		if ((m_base + index) >= m_text->size() ) //index oversized?
			return ' '; //-  return just an empty character..
		else
			return (*m_text)[m_base + index];
	}

protected:
		string	*	m_text;		//the actual text - part of it is the 
								//actual text chunk.
		unsigned    m_base;		//first character of assigned text chunk
		unsigned	m_pos;		//position - number of actual character
		unsigned	m_count;	//actual count of the character in the 
								//text chunk, that has to be drawn
		font	*	m_glyph;	//actual (selected) font's glyph
		double		m_scale;	//requested font size / original m_glyph size
		double		m_anchor;	//0.0: at the start, 0.5: middle, 1.0: at the end
	writing_mode_e	m_wmode;	//text direction
		bool		m_valid;	//true, if the text position has been properly 
								//initialized, and is ready to render the text
		transform	m_local;	//(individual rotation + transformation (dx, dy))
		bool		m_auto_kerning;	//true, when auto-kerning is enabled, false
									//when the value of kerning is constant
		double		m_kerning_value;//value of the constant kerning - distance
									//between the letters.
}; //text_position

}; //namespace svg
#endif SVG_TEXT_POSITION_INCLUDED
