//----------------------------------------------------------------------------
// SVG Viewer - Version 1.0
// A library of SVG renderer
// Copyright (C) 2008 Witold Jaworski
// Contact:	witold.jaworski@airplanes3d.net
//			http://airplanes3d.net
// 
// Contains elements from:
//			AGG - a graphic library (Copyright Maxim Shemanarev, mcseem@antigrain.com)
//          http://antigrain.com
// uses functions from:
//          Expat - a XML parser library (Copyright James Clark)
// 
// SVG Viewer is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
// 
// SVG Viewer is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with SVG Viewer; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, 
// MA 02110-1301, USA.
//----------------------------------------------------------------------------

#ifndef SVG_USE_INCLUDED
#include "svg_shape.h"
#include "svg_req_render.h"
#define SVG_USE_INCLUDED

namespace svg
{
//////////////////////////////////////////////////////////////////////
// use: a reference to another element in the SVG graph
//////////////////////////////////////////////////////////////////////
class use : public shape
{

	public:
		//////////////////////////////////////////////////////////////////////
		// constructor and overriden virtual methods
		//////////////////////////////////////////////////////////////////////

		use(element *parent, unsigned level): shape(parent, level), 
		m_ref(NULL),
		m_x(0), m_y(0),
		m_width(0), m_height(0),
		m_from_xml(true) //it is never changed anymore in this class
		{ 
			m_href = invalid_handle;
		}
	
		//alternative constructor: assigns a known reference
		//Arguments:
		//	href:	id of the referenced element
		use(handle href, element *parent, unsigned level): shape(parent, level) , 
		m_ref(NULL),
		m_x(0), m_y(0),
		m_width(0), m_height(0),
		m_from_xml(false) //it is never changed anymore in this class
		{ 
			m_href = href;
		}

		//Static constructor - used by the parser to create a new reference to an
		//existing element
		//Arguments:
		//	prs:	parser;
		//	parent:	the element, that will be the parent of this element
		//	level:	the level of nesting in the source XML tree (just parent->level()+1 here)
		//	href:	reference, that has to be stored
		static element *create(parser &prs, element *parent, unsigned level, handle href)
		{
			use *elm = new use(href, parent, level);
			return elm;
		}


		virtual const char * tag() {return "use";}
		

		virtual void on_attribute(parser &prs, const char * name, const char *value)
		{
			if (
					(strcmp(name, "xlink:href") == 0) 
			   )
			{	
				m_href  = prs.get_id(value); //id of the referenced element
			}
			else
			if (strcmp(name, "x") == 0) 
			{	
				m_x = prs.to_length(value); //throws an exception for percents
			}
			else
			if (strcmp(name, "y") == 0) 
			{	
				m_y = prs.to_length(value); //throws an exception for percents
			}
			else
			if (strcmp(name, "width") == 0) 
			{	
				m_width = prs.to_length(value); //throws an exception for percents
			}
			else
			if (strcmp(name, "height") == 0) 
			{	
				m_height = prs.to_length(value); //throws an exception for percents
			}
			else
				shape::on_attribute(prs,name,value);
		}//on_attribute

		//request handler:
		virtual handle on_request(request &req) 
		{
			handle result = invalid_handle;
			switch(req.type())
			{
			case request::initialize:	
				result = initialize(req);
				break;

			case request::measure:
			case request::render:
				result = shape::on_request(req);
				break;

			default: //delegate to the reference
				if(m_ref) result = m_ref->on_request(req);
				break;
			}//switch

			if(!m_from_xml && is_valid(result)) 
				return result;
			else
				return skip(req);
		}

	virtual handle render(request &rreq)
	{
		if (!display()) return skip(rreq);
		if (!m_ref) return skip(rreq);

		req_geometry &req = (req_geometry &) rreq;

		//preserve received (bounds) transformation and style
		transform	trans	= req.trans;	
		styleset	style	= req.style;

		//change the request, according styles and transformation of this group
		req.trans.premultiply(m_trans);		
		req.style.update(req,m_attribute_first, m_attribute_last);
	
		handle result = m_ref->on_request(req);

		//restore previous state of the request - (bounds) transformation and style
		req.trans	= trans; 
		req.style	= style;

		//return the result, received from the descendants
		return result;
	}
		
	protected:

	//initializes the content 
	handle initialize(request &req)
	{
		m_ref = req.get_element(m_href);

		m_trans.translate(m_x, m_y);

		if(!m_from_xml && m_ref)
			if (m_id == m_ref->forward_use()) //if it was not touched by anyone...
					return m_ref->on_request(req); //forward this request!
			else 
					return m_ref->skip(req);
		else
			return skip(req);
			
	}//end initialize
	
	protected:
		//////////////////////////////////////////////////////////////////////
		// Instance additional fields
		//////////////////////////////////////////////////////////////////////
		handle	m_href; //handle to the referenced svg::text element
		element *m_ref; //referenced element
		bool	m_from_xml; //true, if this element was defined in the source XML
							//otherwise it is just a placeholder created within
							//parser::store() method to mark the place of an element
							//which was referenced before its definition (and, in
							//effect, receives a lower m_id, than its content)
		double m_x, m_y;	//position of the element

		//width and height are actually not used: (just placed here for the future)
		double m_width;
		double m_height;
}; //use

}; //namespace svg
#endif SVG_USE_INCLUDED
