//----------------------------------------------------------------------------
// svgtex - a SVG vector texture plugin for Blender
// Copyright (C) 2008 Witold Jaworski
// Contact:	witold.jaworski@airplanes3d.net
//			http://airplanes3d.net
//
// This program is based on:	
//			vectex - a SVG vector texture plugin for Blender
//			(Copyright (C) 2008 Marcelo de Gomensoro Malheiros, 
//			 mgmalheiros@gmail.com)
// 
// Contains elements from:
//			AGG - a graphic library 
//			(Copyright Maxim Shemanarev, mcseem@antigrain.com)
//          http://antigrain.com
//
// Uses functions from:
//          Expat - a XML parser library 
//			(Copyright James Clark)
// 
// svgtext is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
// 
// SVG Viewer is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with SVG Viewer; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, 
// MA 02110-1301, USA.
//----------------------------------------------------------------------------

#include <stdio.h>

#include "math.h"

extern "C" // "C" is required by MSVC 6.0 for imports from Blender.exe (plugin.lib)
{
#define PLUGIN_INTERN = 1 /* see blender/extrndef.h */
#include "plugin.h"
}

#include "svgtex.h"
#include "tex_instance.h"

typedef tex::instance	instance;
typedef tex::texel		texel;
typedef tex::tile		tile;
typedef	tex::tileset	tileset;

/*---------------- Plugin management declarations ----------------*/

static char name[] = "svgtex";

/* Names must be less than 15 characters */
#define NR_TYPES 3
static char stnames[NR_TYPES][16] = {"Extend","Repeat","Clip"};

/* varstr and the Cast structure must have the same variables in the same order */ 

static VarStruct varstr[] =
{
	{NUMSLI|INT, "Interpol: ",	1, 0,   4,   "Radius of the pixel interpolation (0 = no interpolation)"},
	{COL|CHA, "Bkgnd",			0, 0, 255,   "Background color"},
	{NUMSLI|INT, "Df.LoD: ",	1, 0,   6,   "Default resolution level (when the OSA is off: 0:256x256,1:512x512,...)"},
	{NUM|INT, "Mem Max",	   20, 0, 200,   "Maximum memory to use (in megabytes, 0=no limit)"},
    {LABEL,   "svgtex 1.0",		0, 0,   0,   ""},
    {TOG|INT, "Enable",			0, 0,   1,   "Enable this texture (disabling frees memory)"},
	{NUMSLI|FLO, "Fid: ",	    0,-5,   0,   "Fidelity of the texture: 0: max, -0.01: classic mipmapping, then worser"},
	{NUMSLI|FLO, "Alpha: ",	    1, 0,   1,	 "Alpha of the background color"},
	{TOG|INT, "Force Df.LoD",   0, 0,   1,   "Always use default image resolution (Df.LoD), regardless of OSA state"},
	{NUM|INT, "Used: ",			0, 0, 999,   "Displays current memory being used (in megabytes)"},
    {LABEL,   "File name",		0, 0,   0,   ""},
    {TEX|CHA, "",				0, 0,  file_name_size-1, "SVG (or a special XML) file name"} 
};
//Beware: the TEXT|CHA field has to be at the end of the varstr array,
//because values from subsequent VarStruct elements would be handled improperly
//by Blender!!!

/* --- Blender reads and WRITE to these two variables here ---*/
//SVGTEX writes to them, also.
//(I feel, that it is not thread-safe, but this is required feature....)

/* Intensity, R, G, B, Alpha, nor.x, nor.y, nor.z - required by old version plugin */
static float result[8];

/* Current frame number */
static float cfra;

/* ----------------- State flags ---------------- */
//(if Blender uses global variables of this module
// - we can introduce our own state flags, because
// we already are not thread - safe....)

/* Error state flag */
static bool in_error = false;

/* True, when Blender has called plugin_init() 
   decreased in plugin_instance_init(), plugin_tex_doit()
*/
static bool  in_init = false; 
/* ---------------------------------------------- */

/*---------------- Plugin management functions ----------------*/
/* Called by Blender only once (during plugin initialization)  */  

int plugin_tex_getversion(void) 
{    
    MSG0("\n\n[plugin_tex_getversion]\n");
    
    return (B_PLUGIN_VERSION);
}

void plugin_getinfo(PluginInfo *info)
{
    MSG0("[plugin_getinfo]\n");

    info->name   = name;
    info->stypes = NR_TYPES;
    info->snames = stnames[0];
    info->varstr = varstr;
    info->nvars  = sizeof(varstr)/sizeof(VarStruct);

    info->result = result;
    info->cfra   = &cfra;

    info->init          = plugin_init;
    info->callback      = plugin_callback;
    info->instance_init = (void (*)(void *)) plugin_instance_init;
    info->tex_doit      = (TexDoit) plugin_tex_doit;
}

/*--- Function not needed by this plugin, but required by Blender: ---*/
/* called when plugin is loaded OR duplicated (by selecting Single User option)*/
void plugin_init(void)
{
    MSG0("[plugin_init]\n");
	in_init = true; 
}

void plugin_callback(int button) 
{
    MSG0("[plugin_callback]\n");
	button = 0; //prevent unused variable warining calls
	in_error = false; //let's hope, that user removes the error reason,
					  //changing this control parameter
}
/*------------------------ Helper functions --------------------------*/

//Handles the catched runtime errors, which have not came from the SVG processor:
//Fills the result[] with proper data and returns a flag which is expected by 
//plugin_tex_doit() procedure
//Arguments:
//	cast:	     plugin control data
//	result:		 non-null when function is called inside plugin_tex_doit()
//				 (the pointer to the result data)
static int on_runtime_error(Cast *cast, float *result = NULL )
{
		perror("SVGTEX: a system exception ocurred=> ");
		cast->enabled = 0;
		in_error = true;
		if(result) return texel::red().copy_to(result);
				   else return 1;
}

//Handles the catched runtime errors, which have came from the SVG processor:
//Fills the result[] with proper data and returns a flag which is expected 
//by plugin_tex_doit() procedure
//Arguments:
//	cast:		 plugin control data
//	e:			 an object, which describes the SVG exception
//	result:		 non-null when function is called inside plugin_tex_doit()
//				 (the pointer to the result data)
static int on_runtime_error(Cast *cast, agg::svg::exception& e, 
													float *result = NULL)
{
    printf("SVGTEX: an SVG processing exception occured=> %s\n",e.msg());
	cast->enabled = 0;
	in_error = true;
	if(result) return texel::red().copy_to(result);
			   else return 1;
}
/*------------------------- Main functions ---------------------------*/

void plugin_instance_init(Cast *cast)
{
    MSG0("[plugin_instance_init]\n");
    try
	{
		if (cast->enabled)
		{
			/* Plugin is enabled, so create instance data */
			MSG0("- plugin started enabled\n");
			instance::create_for(cast);
			in_init = false;
		}
		else
		{
			/* Called by a blender saved file; initialize to disabled state */
			MSG0("- plugin started disabled\n");
			//cast->instance_data is null, but just to be sure:
			instance::discard_from(cast); 
		}
	}//try
    catch(agg::svg::exception& e)
    {
        on_runtime_error(cast, e); 
    }
	catch(...)
	{
        on_runtime_error(cast); 
	}

}

/*------------------------------------------------*/
int plugin_tex_doit(int stype, Cast *cast, float *texvec, float *dxt, float *dyt, float *result)
{
	//Before returning the result, we have to examine many things:
    try
    {
		//TEST 1: is this plugin instance enabled?
		if(!cast->enabled)
		{
			if(cast->instance_data != NULL)
			{
				MSG0("- plugin deinitialized\n");
				instance::discard_from(cast);
			}

			if(in_error) //set in catch() statement to true
			{			 //reset by changing of any button to false
				return texel::red().copy_to(result);
			}
			else
			{
				return texel::transparent_black().copy_to(result);
			}
		}

		//TEST 2: has the instance object been created?
		//(in_init = true also for the case, when the plugin texture 
		//instance has been duplicated, in effect of user selection
		//of SingleUser option: Blender does not call plugin_instance_init()
		//in such case!)
		if(cast->instance_data == NULL || in_init == true)
		{
			MSG0("- plugin reinitialized\n");
			instance::create_for(cast);
			in_init = false;
		}

		instance *inst = (instance *) cast->instance_data;

		//TEST 3: Are the last used file name and the actual typed one 
		//		  identical?
		if(strcmp(inst->file_name(),cast->file_name) != 0)
		{
			MSG1("- plugin reset (file name changed to '%s')\n", cast->file_name);
			instance::discard_from(cast);
			inst = instance::create_for(cast);
		}
		
		//TEST 4: Is the instance object enabled? (in case of problems with
		//		  parsing SVG file it is disabled)
		if(!inst->is_enabled())
		{
			return texel::red().copy_to(result);
		}
		
		//TEST 5: Are the last used background color and the actual one 
		//		  identical?
		agg::rgba8 c2 = cast->bkgnd_color();
		if(! svg::is_equal(inst->background(), c2))
		{
			MSG0("- tiles reset (background color changed)\n");
			inst->tiles().reset();
			inst->background(c2);
			//Let's reset whole plugin instance, also for this reason -
			//like for the file name change. The time of parsing anew
			//the SVG file is short enough to be acceptable -
			//usually the rendering takes much more time...
			// instance::discard_from(cast);
			//inst = instance::create_for(cast);
		}

	//OK, we have passed tests. Now accept the parameters, that may be changed:

		//accept proposed default LOD:
		if(cast->def_LOD != inst->default_LOD())
		{
			inst->default_LOD(cast->def_LOD);
		}

		//Have we force default resolution (LOD), regardless of requested one?
		if (cast->fix_def_LOD == 1) //yes!
		{	
			dxt = dyt = NULL; //for the rest of the code it will looks like 
							  //Blender OSA has been disabled (dxt,dyt are NULL, then)
		}

	//Calcualte the result:
		//(the most important procedure:)
		texel p = inst->sample((instance::extend_mode) stype,texvec[0],texvec[1],
											dxt,dyt, cast->interpol, cast->fidelity);

	//Final clean-up:
		//Refresh the actual memory situation:
		cast->mem = inst->tiles().used_memory(); //in MB 

		//tileset maintenence:
		if(cast->mem_max != 0 && cast->mem > cast->mem_max)
		{
			inst->tiles().limit(cast->mem_max);

			cast->mem = inst->tiles().used_memory();
		}
		
		//Finally - return the result!
		return p.copy_to(result);

	}//try
    catch(agg::svg::exception& e)
    {
        return on_runtime_error(cast, e, result); 
    }
#ifndef DEBUG //catching up the errors is useful for the debugging....
	catch(...)
	{
        return on_runtime_error(cast, result); 
	}
#endif 
}

