//----------------------------------------------------------------------------
// svgtex - a SVG vector texture plugin for Blender
// Copyright (C) 2008 Witold Jaworski
// Contact:	witold.jaworski@airplanes3d.net
//			http://airplanes3d.net
//
// This program is based on:	
//			vectex - a SVG vector texture plugin for Blender
//			(Copyright (C) 2008 Marcelo de Gomensoro Malheiros, 
//			 mgmalheiros@gmail.com)
// 
// Contains elements from:
//			AGG - a graphic library 
//			(Copyright Maxim Shemanarev, mcseem@antigrain.com)
//          http://antigrain.com
//
// Uses functions from:
//          Expat - a XML parser library 
//			(Copyright James Clark)
// 
// svgtext is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
// 
// SVG Viewer is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with SVG Viewer; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, 
// MA 02110-1301, USA.
//----------------------------------------------------------------------------
#include "svgtex.h"
#include "tex_instance.h"

namespace tex
{
	//Constructor
	//Arguments:
	//	cast:	the data structure with parameters, received from Blender
	instance::instance(Cast *cast)
	{
		m_background = cast->bkgnd_color();
		default_LOD(cast->def_LOD);
		memset(m_file_name,0,sizeof(m_file_name));
		
		//try to load the SVG file and prepare the SVG graph for rendering:
		if(strcmp(cast->file_name,"") != 0)
		{
			strncpy((char *) file_name(), cast->file_name, sizeof(m_file_name)-1);
			m_svg_graph.load(file_name());
			m_enabled = (m_svg_graph.elements() > 0);

#ifdef DEBUG
			if (m_enabled) 
			{MSG1("- instance init: file '%s' loaded sucessfully.\n",file_name());}
			else  {MSG1("- instance init: failed to load file '%s'.\n",file_name());}
#endif DEBUG

		}
		else
		{
			MSG0("- instance init: no file name\n");
			m_enabled = false;
		}
	}

	//Renders a tileset image (representd as <pixf>)
	//Arguments:
	//	level:	level of details of the image (total resolution = 2^level)
	//	row:	vertical position of the tile on the picture;
	//	col:	horizontal position of the tile on the picture;
	void instance::render(int32 level, int32u col, int32u row, pixfmt &pixf)
	{
		double		resolution = double(1 << level);
		double		width	= m_svg_graph.width();
		double		height	= m_svg_graph.height();
		//double		size = (width > height) ? width: height;
		transform	mtx;

		rectangle	clip	= m_svg_graph.extents();

		//move center of the image to (0,0)
		mtx.translate(-(clip.x1 + clip.x2)*0.5, -(clip.y1 + clip.y2)*0.5);

		//scale to stretch rectangle to a square: 
		//(Blender does the same with static images):
		mtx.scale(resolution/width, resolution/height); 
		
		//move back to center of the image
		mtx.translate(resolution * 0.5, resolution * 0.5);

		//move to render selected tile: 
		mtx.translate(	- tile_size * int(col), - tile_size * int(row));

		//add tile border adjustment:
		mtx.translate(tile_border, tile_border);
		
		//Get actual clipping coordinates:
		mtx.transform(&(clip.x1), &(clip.y1));
		mtx.transform(&(clip.x2), &(clip.y2));

		//increase the clipping area by tile_border pixels in each direction: 
		//(a margin for the pixel interpolation)
		clip.x1 -= tile_border;
		clip.y1 -= tile_border;
		clip.x2 += tile_border;
		clip.y2 += tile_border;

		//img: size of the image buffer:
		rectangle img = rectangle(0,0,double(pixf.width()),double(pixf.height()));
		//restrict clip area 
		clip = agg::intersect_rectangles(clip,img);

		m_svg_graph.render(pixf, clip, mtx, m_background, 1.0);
	}
	
	//Helper function: for a value find nearest power of 2:
	//Argument:
	//	val:	a value, which is compared to the powers of 2
	static int32u nearest_power_of_2(int32u val)
	{
		int32u power;
		for (power = 1; val >> power; power ++);
		return power;
	}

	//Returns a texel for given texture coordinates
	//Arguments:
	//	mode:	 selected mode of texture extension	(expand, clip, repeat)
	//	u,v:	 the texture coordinates, received from Blender (negaitive or 
	//			 positive, range -1.0 .. 1.0 matches the picture size)
	//	dxt,dyt: optional: two vectors (dxt[0],dyt[0]), (dxt[1],dyt[1]) 
	//			 which describe the size of teh pixel. NULL when Blender's OSA 
	//			 is off. In the same texture coordinate space like u,v 
	//	radius:	 radius of the pixel interpolation (0 = no pixel interpolation)
	//	fidelity:a value  from -5 to 0, describing eventual degradation of
	//			 the image.
	texel instance::sample(instance::extend_mode mode, float u, float v, 
						float *dxt, float *dyt, int32u radius, float fidelity)
	{
		if (!clamp_coord(mode, &u) || !clamp_coord(mode, &v))
		{
			return texel::transparent_black();
		}
		else
		{
			//1. Determine the desired LOD (i.e. resolution):
			int32u ideal_res; // ideal resolution,
			int32u level;	  // normal level of details (its resolution > ideal_res)
			int32u flevel;	  // sublevel (its resolution < ideal_res) - used when 
							  // fidelity < 0.0		
			int32u weight;	  // weight (subpixel scale) of the texel on the LoD = level
							  // (i.e. the higher LoD) - used in fidelity and mipmap
							  // calculations.
			texel  result;	  // result of this caluclation

			if (dxt && dyt)
			{
				double signed_area = dxt[0]*dyt[1] - dxt[1]*dyt[0];
				double pixel_size  = sqrt(fabs(signed_area));
				ideal_res = int32u(2.0/pixel_size);
				level = nearest_power_of_2(ideal_res);

				//do not increase the max. resolution suddenly for a 
				//degenerated case:
				if(level > tiles().max_level())
				{
					double dot_product = dxt[0]*dyt[0] + dxt[1]*dyt[1];

					//degenerated case: pixel is at angle < 6 degrees:
					if (dot_product != 0.0 && fabs(signed_area/dot_product) < 0.1)
					{
						MSG0("*");
						//use the maximum available resolution:
						level = tiles().max_level();
						ideal_res = 1 << level;
					}
				}
			}
			else //use default resolution
			{
				ideal_res = default_resolution();
				level = m_default_LOD;
			}//end dxt && dyt


			//Does the calcualted LoD need to be alerted by the fidelity setting?
			if (fidelity < 0.0) 
			{
				int32  dfid = int(floor(fidelity)); //notice, that fidelity <= 0.0,
												    //so for -2.8 floor() returns  -3.0!
				flevel = int32u(int(level) + dfid);	//dfid is always negative!
				//weight of the texel from the igher resolution (dfid is always negative):
				weight = int32u(int32((dfid - fidelity) * texel::sub_scale) / dfid); 
			}
			else
			{
				weight = texel::sub_scale; //conceptually: 1.0 
			}

			//2. calculate the texel:			
			if((flevel < level) && (flevel >= 0)) //no degenerated LoD, nor idenitity mipmapping
			{

				//When mipmapping is on:
				int32u gt_res = 1 << level;
				int32u lt_res = 1 << flevel;

				// Conceptually: weight *= (ideal_res - lt_res) / (gt_res - lt_res):
				weight *= ((ideal_res - lt_res)*texel::sub_scale/(gt_res - lt_res));
				weight >>= texel::sub_shift;

				//when weight of the "second side" is not a zero: mipmap it!
				if (weight < texel::sub_scale) // weight of the lower level is nonzero:
				{
					result = tiles().sample(u, v, flevel, radius, *this);

					if (weight > 0) // weight of the higher level is important?
					{
						//mip: the texel from the higher resolution:
						texel mip = tiles().sample(u, v, level, radius, *this);
						//... to mix them together:
						result.mix(mip,weight); 
					}
				}
				else //no need to use the lower resolution values for this weight
				{
					result = tiles().sample(u, v, level, radius, *this);
				}
			}
			else //no need to use the lower resolution values, at all!
			{
				result = tiles().sample(u, v, level, radius, *this);
			}
			
			//3. That's all: return the result.
			return result;

		}//clamp_cooords = true
	}

}//end namespace tex
