//----------------------------------------------------------------------------
// svgtex - a SVG vector texture plugin for Blender
// Copyright (C) 2008 Witold Jaworski
// Contact:	witold.jaworski@airplanes3d.net
//			http://airplanes3d.net
//
// This program is based on:	
//			vectex - a SVG vector texture plugin for Blender
//			(Copyright (C) 2008 Marcelo de Gomensoro Malheiros, 
//			 mgmalheiros@gmail.com)
// 
// Contains elements from:
//			AGG - a graphic library 
//			(Copyright Maxim Shemanarev, mcseem@antigrain.com)
//          http://antigrain.com
//
// Uses functions from:
//          Expat - a XML parser library 
//			(Copyright James Clark)
// 
// svgtext is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
// 
// SVG Viewer is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with SVG Viewer; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, 
// MA 02110-1301, USA.
//----------------------------------------------------------------------------
#ifndef TEX_TEXEL_INCLUDED
#include <stdio.h>
#include "svg_basics.h"

extern "C" // "C" is required by MSVC 6.0 for imports from Blender.exe (plugin.lib)
{
#define PLUGIN_INTERN = 1 /* see blender/extrndef.h */
#include "plugin.h"
}

#define TEX_TEXEL_INCLUDED
namespace tex
{
	typedef agg::int8u				int8u;
	typedef agg::int32u				int32u;
	typedef agg::int32				int32;
	typedef agg::rendering_buffer	rndbuf;
	typedef svg::rectangle			rectangle;
	typedef svg::transform			transform;

//////////////////////////////////////////////////////////////////////
// texel:	the basic (conceptual) element of the texture. 
// Contains not only the color, but also information about 
// the intensity and the direction of the normals.
//////////////////////////////////////////////////////////////////////
struct texel
{
public:
	//texel calculations are conducted with precision
    enum sub_scale_e
    {
        sub_shift = 14,           
        sub_scale = 1 << sub_shift,   
		sub_halfscale = sub_scale >> 1, //used in the slope calc optimisation
        sub_mask  = sub_scale - 1,     
		char_shift = sub_shift - 8 //a shift for the character values
    };
//---------------------------- fields -------------------------------
	//all values expressed as <sub_shift> values (i.e. 1.0 = sub_mask)

	int32u i; //intensity
	int32u r; //red:			
	int32u g; //green:
	int32u b; //blue:
	int32u a; //alpha

	int32 dx; //deviation of the normal vector in the x (or u) direction
			    //(from - sub_mask to + sub_mask)
	int32 dy; //deviation of the normal vector in the y (or v) direction
		    	//(from - sub_mask to + sub_mask)
public:
//---------------------------- methods ------------------------------
	//Simplest constructor
	texel()
	{
		i = r = g = b = a = dx = dy = 0;
	}

	//Straightforward constructor
	texel(int32u i_, int32 r_, int32 g_, int32u b_, int32u a_, int32 dx_ = 0, int32 dy_ = 0)
	{
		i = i_; //"::" prefix, because there are functions of the same names
		r = r_;
		g = g_;
		b = b_;
		a = a_;
		dx = dx_;
		dy = dy_;
	}
	
	//Creates a texel from array of four RGBA values
	//Arguments:
	//	p:	the array of the pixel values (it is assumed to have the pixfmt 
	//		channel order)
	texel(int8u *p)
	{
		r = from_byte(p[pixfmt::order_type::R]);
		g = from_byte(p[pixfmt::order_type::G]);
		b = from_byte(p[pixfmt::order_type::B]);
		a = from_byte(p[pixfmt::order_type::A]);
		i = intensity(r, g, b, a);
		dx = 0;
		dy = 0;
	}

	//Creates a texel from a color structure
	//Arguments:
	//	c:	the color structure
	texel(color &c)
	{
		r = from_byte(c.r);
		g = from_byte(c.g);
		b = from_byte(c.b);
		a = from_byte(c.a);
		i = intensity(r, g, b, a);
		dx = 0;
		dy = 0;
	}

	//Creates a texel from given pixel of the PixelFormat data
	//Arguments:
	//	pf:	 the pixel buffer
	//	x,y: position of the pixel within the pixel buffer
	//	slope:	true, when the dx,dy slope has to be calculated
	//It is the only one constructor, that can return a texel
	//with non-zero m_dx, m_dy.
	texel(pixfmt &pf, int32 x, int32 y, bool slope = false);

	/*----------------------- Static colors --------------------------- */
	static texel transparent_black()
	{
		color c = color(0,0,0,0);
		return texel(c);
	}

	static texel red()
	{
		color c = color(255,0,0);
		return texel(c);
	}
	
	static texel black()
	{
		color c = color(0,0,0);
		return texel(c);
	}

	static texel white()
	{
		color c = color(color::base_mask,color::base_mask,color::base_mask);
		return texel(c);
	}
	/*----------------------- Other methods --------------------------- */
	//Helper function to write texel to the Blender result buffer
	//Arguments:
	//	result:	the address of Blender result buffer
	//Returns: the flags, which should return the tex_doit() -
	//just place a "return a_texel.copy_to(result);" at its end.
	int copy_to(float *result)
	{
		if (a < sub_mask && a > 0) //alpha between 0.0 and 1.0: let's premultiply
		{
			result[1] =  (float (r*a>>sub_shift) / sub_scale);
			result[2] =  (float (g*a>>sub_shift) / sub_scale);
			result[3] =  (float (b*a>>sub_shift) / sub_scale);
		}
		else //no premultiplication needed!
		{
			result[1] =  (float (r) / sub_scale);
			result[2] =  (float (g) / sub_scale);
			result[3] =  (float (b) / sub_scale);
		}
		result[4] =  (float (a) / sub_scale);
		result[0] =  (float (i) / sub_scale) * result[4]; //intensity * alpha

		result[5] = (float (dx) / (sub_scale));
		result[6] = (float (dy) / (sub_scale));
		
		result[7] = 0;
		return (TEX_INT | TEX_RGB | TEX_NOR);
	}

	
	//Mixes this texel with another one, with given weight
	//Arguments:
	//	p:		another texel, which is mixed
	//	weight:	share of p in the result (subscaled to integrer,
	//			conceptually in 0.0 ... 1.0)
	void mix(texel &p, int32u weight);

	//Returns a texel, which is the result of averaging of the pixels 
	//around x,y, within given radius.
	//Arguments:
	//	radius:	the level of the averaging: radius of the area,
	//			containing texels to be averaged. 0 returns just
	//			a texel corresponding to the pixel at x,y.
	//	x,y:	coordinates of the source pixel. They can be a "big"
	//			values, exceeding the <pf> dimensions. It is assummed,
	//			that the <pf> describes a fragment of this large picture
	//			build from many similar pixel formats of identical size.
	//			Another assumption - <pf> is the proper one, i.e. fragment
	//			which conatins the pixel corresponing to x,y on the large
	//			picture.
	//	pf:		pixel format - a buffer with the pixels
	//	tu,tv	pixel coordinates, corresponding to x, y. They can differ
	//			from x,y values by less than 1.0, never more!
	static texel sample(int32u radius, int32 x, int32 y, pixfmt &pf, 
															float tu, float tv);

protected:
	inline int32u static from_byte(int8u c) {return ((int32u) c) << char_shift;}

	//Encapculates the method of calculating texel intensity.
	//Actually it does not use the "human" intensity, just the
	//simple average of all color channel values, decreased by the alpha
	//All values are presumed to
	inline int32u static intensity(int32u r, int32u g, int32u b, int32u a)
	{
		return (r + g + b) / 3; //(((r + g + b) / 3) * a ) >> sub_shift; 
	}
	
	//The same method, but for a "raw pixel" data, as available 
	//from a pixel format.
	inline int32u static intensity(int8u *p)
	{
		return intensity(
							from_byte(p[pixfmt::order_type::R]),
							from_byte(p[pixfmt::order_type::G]),
							from_byte(p[pixfmt::order_type::B]),
							from_byte(p[pixfmt::order_type::A])
						);
	}

	//Helper method, that calculates a weighted slopes between two intensities
	//Arguments:
	//	tl,tr,bl,br:				intensity values (in subpixel accuracy)
	//								of four neigbour pixes (TopLeft, TopRight,
	//								BottomLeft, BottomRight)
	//	dx, dy:						result slope of intensities in the u and v drection
	//	u_opposite, u_ratio:		contribution of the left and the right pixels
	//								to the effecitive pixel intensities 
	//								(notice that following rule:
	//								 u_opposite + u_ratio = sub_scale
	//								 should be always true!)
	//	v_opposite, v_ratio:		contribution of the top and the bottom pixels
	//								to the effecitive pixel intensities 
	//Fills the slope in the u (dx) and v (dy) direction as the weighted average
	static inline void calc_slope(int32u tl, int32u tr, int32u bl, int32u br, 
									int32 *dx, int32 *dy,
									int32u u_opposite = texel::sub_halfscale, 
									int32u u_ratio = texel::sub_halfscale,
									int32u v_opposite = texel::sub_halfscale, 
									int32u v_ratio = texel::sub_halfscale
								  ) 
	{

		int32 tdx = int(tr - tl); //upper change of U
		int32 bdx = int(br - bl); //lower change of U
		int32 ldy = int(tl - bl); //left change of V (reverted direction, V grows up)
		int32 rdy = int(tr - br); //right change of V (reverted direction, V grows up)
		
		if (tdx == 0 && bdx == 0 && ldy == 0 && rdy == 0) //nothing to do
		{
			*dx = *dy = 0;
		}
		else
		{
			if (u_ratio == sub_halfscale && v_ratio == sub_halfscale)
			{	//if the result pixel is exactly in the middle of the four components:
				*dx = (tdx + bdx)/2;
				*dy = (ldy + rdy)/2;
			}
			else //full, weighted version
			{
				*dx = (tdx * int(v_opposite) + bdx * int(v_ratio))/sub_scale;
				*dy = (ldy * int(u_opposite) + rdy * int(u_ratio))/sub_scale;
			}
		}
	}// calc_slope
	
	//Helper function, used in the filter calculations: 
	//fills four neighbour texels with data
	//Arguments:
	//	radius:	the level of the averaging: radius of the area,
	//			containing texels to be averaged. 0 returns just
	//			a texel corresponding to the pixel at x,y.
	//	x,y:	coordinates of the source pixel in the pixel format <pf>
	//	pf:		pixel format - a buffer with the pixels
	//	p_tl:	top-left neighbour - to be filled with data
	//	p_tr:	top-right neighbour - to be filled with data
	//	p_bl:	bottom-left neighbour - to be filled with data
	//	p_br:	bottom-right neighbour - to be filled with data
	//One of the p_* texels is the texel corresponding to x,y pixel:
	//- when <radius>is odd, it is <p_tl>, when even - <p_br>
	static void get_neighbours(int32u radius, int x, int y, pixfmt &pf,
						texel &p_tl, texel &p_tr, texel &p_bl, texel &p_br);

	//Helper function, used in the filter calculations.
	//Calculates a texel of u,v coordinates exactly in the center of x,y pixel
	//Arguments:
	//	radius:	the level of the averaging: radius of the area,
	//			containing texels to be averaged. 0 returns just
	//			a texel corresponding to the pixel at x,y.
	//	x,y:	coordinates of the source pixel in the pixel format <pf>
	//	pf:		pixel format - a buffer with the pixels
	static texel sample(int32u radius, int x, int y, pixfmt &pf);

	private:
	//helper function for calculating the averages (see texel::filter() code):
	//l,r:		weighted values ("left", "right")
	//ratio:	share of the <r> value	(conceptually: weight), in subpixel.			
	//opposite: conceptually: (1-weight) in subpixel.

	static inline int32u w_avg(int32u l, int32 r, int32u ratio,int32u opposite)
	{
		return (l * opposite + r * ratio) >> sub_shift;
	}
	//helper function for calculating the averages (see texel::filter() code):
	//tl, tr, bl, br:	values belonging to four neigbour texels
	//u_ratio,v_ratio:	share (scaled up by sub_scale!) of the bottom - right 
	//					texel
	//u_opposite,v_opposite:	share (scaled up by sub_scale!) of the 
	//							upper - left texel
	static inline int32u w_avg(int32u tl, int32 tr, int32 bl, int32 br, 
										int32u u_ratio, int32u v_ratio, 
										int32u u_opposite, int32u v_opposite)
	{
		return w_avg(
						w_avg(tl,tr,u_ratio,u_opposite),  
						w_avg(bl,br,u_ratio,u_opposite), 
						v_ratio,
						v_opposite
					); 
	}

};//end texel

}//end namespace tex
#endif TEX_TEXEL_INCLUDED
