//----------------------------------------------------------------------------
// svgtex - a SVG vector texture plugin for Blender
// Copyright (C) 2008 Witold Jaworski
// Contact:	witold.jaworski@airplanes3d.net
//			http://airplanes3d.net
//
// This program is based on:	
//			vectex - a SVG vector texture plugin for Blender
//			(Copyright (C) 2008 Marcelo de Gomensoro Malheiros, 
//			 mgmalheiros@gmail.com)
// 
// Contains elements from:
//			AGG - a graphic library 
//			(Copyright Maxim Shemanarev, mcseem@antigrain.com)
//          http://antigrain.com
//
// Uses functions from:
//          Expat - a XML parser library 
//			(Copyright James Clark)
// 
// svgtext is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
// 
// SVG Viewer is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU General Public License
// along with SVG Viewer; if not, write to the Free Software
// Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, 
// MA 02110-1301, USA.
//----------------------------------------------------------------------------
#ifndef TEX_TILE_INCLUDED
#include "tex_texel.h"
#define TEX_TILE_INCLUDED
namespace tex
{
//////////////////////////////////////////////////////////////////////
// tile:	a square fragment of the image.
// For small images,  (< tile_size) contains whole image, for larger 
// ones - its part, only.
// A tile contains the image buffer, rendering buffer and pixel format
// Rendered image is always bigger than the nominal size (represented
// by the pixel format. It contains a border around its nominal size,
// a <tile_border> pixels wide. This border is needed when texel in-
// terpolation is used to obtain a texel(u,v) value.
//
// Tiles are organized in a collection, which contains all mipmaps
// of the texture. Such collection is implemented by tex::tileset class.
//////////////////////////////////////////////////////////////////////

//forward declaration, needed for this class:
#ifndef TEX_INSTANCE_INCLUDED
class instance;
#endif TEX_INSTANCE_INCLUDED

class tile
{
public:
	enum constants
	{
		no_aging	= -1,	//m_age of the constant color tile uses it 
		unused		= 0,	//for empty tiles
		initial_age = 1,	//for rendered tiles
	};

protected: //data fields:

	int	   m_age;	//counter, used by the tileset for memory management 
					//(It periodically removes the tiles least used, in 
					//the tileset::limit() method).
					//The tileset object has a timestamp counter, which 
					//increases at every request for a texel (every call
					//to plugin_tex_doit()). m_age of a tile is always 
					//set to actual tileset::timestamp value, when a texel 
					//from this tile was used. This means, that the tiles
					//often used have greater m_age, than the tiles seldom
					//used.
	int8u  m_mono[bytes_pp]; //for handling a special case, when whole image
							 //contains a constant color: this is this color.
							 //It will be used instead of m_image memory 
							 //pointer.
	int8u *m_image;	//image buffer, used by m_rbuf and m_pixf: it contains
					//always a picture enlarged, comparing to nominal size,
					//by the <tile_border> pixels from each side. For unini-
					//tialized tiles it is NULL. For constant color tiles it
					//contains the address of m_mono array.
	rndbuf m_rbuf;	//rendering buffer:
	pixfmt m_pixf;	//pixel format: declared for nominal width and height
					//(in fact its buffer contains more pixels - on the border)
public:
	tile():
		m_image(NULL),
		m_pixf(m_rbuf),
		m_age(unused)
	{
		memset(m_mono,0,sizeof(m_mono)); //clear the mono array
	}

	~tile() { reset();}

	//Initializes this object - second phase of construction!
	//Arguments:
	//	inst:	the instance object - it contains the tileset, which
	//			contains this tile. Needed to render the image.
	//	level:	resolution level (a power of 2 - for example 8 means
	//			an image of 256x256 pixels. Higher levels implies
	//			multi-tile image (where row and col can be > 0)
	//	col:	for single-tile images (up to tile size): 0
	//			for multi-tile images: column of the tile on the
	//			picture (0 is the rightmost one)
	//	row:	for single-tile images (up to tile size): 0
	//			for multi-tile images: row of the tile on the
	//			picture (0 is the topmost one)
	void initialize(instance &inst, int32u level, int32u col, int32u row);

	//Resets tile to the initial (unused) state
	void reset();

	//Some helper functions:
	//returns true, when this tile is used 
	inline bool is_enabled() {return m_age != unused;}

	//returns true, when this is a solid color tile:
	inline bool is_constant_color() {return m_image == m_mono;}
	
	//returns true, when this tile has been initialized:
	inline bool is_initialized() {return m_image != NULL;}

	//Accessors to the m_age field:
	int age() {return m_age;}
	void age(int val) {m_age = val;}
	
	//Returns a reference to this tile, changes its m_age value
	//Arguemnts:
	//	age: the new value - actual tieset::timestamp()
	tile &reference(int32u age)
	{
		if (!is_constant_color()) m_age = age;
		return *this;
	}

	//Returns a texel from a tile.
	//Arguments:
	//	tu,tv:		pixel coordinates (expressed up to fractional part of
	//				the pixel) - in range from -0.5 up to <resolution> - 0.5;
	//	resolution: resolution of the whole picture, which this
	//				tile belongs to. A tile can contain a fragment of this 
	//				picture, only. This fragment contains the corresponding 
	//				(tu,tv) pixel, anyway.
	//	radius:		span of the pixel interpolation (pixels):
	//				0 = no interpolation.
	texel sample(float tu, float tv, int32u resolution, int32u radius)
	{
		if (is_constant_color())
		{
			return texel(m_mono);
		}
		else
		{
			//conversion tu, tv to integers in this call reduces their fractal
			//part: -0.5 becomes 0, and 2047.5 becomes 2047. This effect
			//is desired for the tex::filter() function, which expects the
			//integer pixel coordinates.
			return texel::sample(radius, int(tu), int(tv), m_pixf,tu, tv);
		}
	}//end sample

	//verifies the image buffer, checking if it contains identical pixels
	//returns true, when all pixels are identical
	bool is_mono()
	{
		int offs, size = m_rbuf.height()*m_rbuf.stride();
		for (offs = bytes_pp; offs < size; offs+=bytes_pp)
		{
			if(memcmp(m_image, m_image+offs, bytes_pp) != 0) break; 
		}

		return offs == size;
	}

	//Returns amount of the image memory, allocated to this tile
	int32u memory()
	{
		if(is_enabled() && !is_constant_color())
		{
			return m_rbuf.width()*m_rbuf.height()*bytes_pp;
		}
		else
		{
			return 0;
		}
	}

protected:
	//Allocates an image on the Blender heap.
	//Arguments:
	//	resolution:	size of the edge of image square, in pixels
	int8u *create_image(int32u resolution)
	{
		return (int8u *) mallocN(resolution*resolution*bytes_pp,"svgtex::image");
	}

	//Releases memory from the Blender heap
	//Arguments:
	//	image:	image to be released (just for code clarity)
	//Always return NULL (to allow the m_image to be reinitialized to NULL)
	int8u *release_image(int8u *image)
	{
		if(image) freeN(image);
		return NULL;
	}
};//end tile

}//end namespace tex
#endif TEX_TILE_INCLUDED
